"""EPREM runtime parameter arguments, definitions, and metadata.

This module includes the following objects::
* `~parameters.BaseTypesH` provides access to all EPREM runtime constants (see
  src/baseTypes.h), given a path to the directory containing EPREM source code.
* `~parameters.ConfigurationC` provides definitions and default arguments of all
  parameters relevant to a simulation run (see src/configuration.c), given a
  path to the directory containing EPREM source code.
* `~parameters.ConfigFile` represents the contents of a named configuration file
  that contains user-provided arguments for a subset of simulation parameters,
  given a path to the file.
* `~parameters.Runtime` provides a unified interface to all parameter arguments
  used in a particular simulation run, given appropriate paths.
* `~parameters.Arguments` is an instance of `aliased.Mapping` that
  supports aliased access to simulation and post-processing parameters, given an
  instance of `~parameters.Runtime`.

Notes
=====

Metadata
--------
* `~parameters.Runtime` represents the full set of parameters relevant to a
  particular simulation run in two regards: 1) the EPREM distribution (e.g.,
  'epicMas' or 'epicEnlil'), and 2) the combination of user-provided and default
  arguments. There is a long-term goal to develop a single modular EPREM
  distribution, completion of which would remove the first item.
* `~parameters._CONFIGURATION_C` contains metadata for a subset of the
  parameters that `~parameters.Runtime` represents; `~parameters.Runtime` is
  always the canonical collection.
* `~parameters._BASETYPES_H` is always a proper subset of `~parameters.Runtime`.
* `~parameters.Runtime` + `~parameters._LOCAL` represents the full set of
  parameters available to post-processing code in this package.

Terminology
-----------
The word "parameter" can have multiple meanings in the contexts to which this
module is relevant. The python built-in module `inspect` uses "parameters" to
mean the names representing positional and keyword arguments. This is consistent
with formal definitions of a function parameter (cf.
https://developer.mozilla.org/en-US/docs/Glossary/Parameter). It is also common
to refer to config-file options as parameters, in a way that is similar but not
identical to the formal software defintion above. This module attempts to
respect the formal distinction between parameters and arguments.
"""

import abc
import argparse
import functools
import numbers
import pathlib
import re
import typing
import json

from goats.core import algebra
from goats.core import aliased
from goats.core import iotools
from goats.core import iterables
from goats.core import numerical
from goats.core import quantities


class BaseTypeDef:
    """Pattern parser for defined constants in baseTypes.h."""

    @property
    def pattern(self):
        return re.compile(r"""
            # the start of the string (no following whitespace)
            \A
            # the `#define` pre-processor directive
            \#define
            # at least one whitespace character
            \s+
            # the defined name
            (?P<name>[A-Z_]+)
            # at least one whitespace character
            \s+
            # the defined value
            (?P<value>[-+\.\*\/\d\w\s\(\)]+)
            # the end of the string (no trailing whitespace)
            \Z
        """, re.VERBOSE)

    def match(self, line: str):
        """Identify lines that define a runtime constant."""
        if match := self.pattern.match(line.strip()):
            return match.groupdict()

    def parse(self, parsable: typing.Dict[str, str]):
        """Parse a line that defines a runtime constant."""
        return parsable['name'], parsable['value']


class SourceFile(iterables.MappingBase):
    """An object representing parameters in an EPREM source file."""

    def __init__(
        self,
        db_key: str,
        filename: str,
        source: typing.Union[str, pathlib.Path]=None,
    ) -> None:
        """Initialize an instance of this class."""
        path = self.build_path(filename, source)
        if path and path.exists():
            loaded = self.load_from_source(iotools.TextFile(path))
        else:
            loaded = self.load_from_package(db_key)
        self.definitions = self.standardize(loaded)
        """The parameter definitions in this file."""
        super().__init__(tuple(self.definitions))

    @abc.abstractmethod
    def standardize(self, loaded: dict):
        """Convert loaded attributes into standard definitions.
        
        This abstract method provides a default implementation that simply
        returns the input object.
        """
        return loaded

    def build_path(
        self,
        filename: str,
        source: iotools.PathLike=None,
    ) -> typing.Optional[pathlib.Path]:
        """Create the full path to the source file, if possible."""
        try:
            path = iotools.ReadOnlyPath(source)
            if path.is_dir():
                path /= filename
        except TypeError:
            path = None
        return path

    def load_from_package(self, key: str) -> dict:
        """Load argument definitions from the package defaults."""
        path = pathlib.Path(__file__).with_suffix('.json')
        with pathlib.Path(path).open('r') as fp:
            loaded = json.load(fp)
        return loaded[key]

    @abc.abstractmethod
    def load_from_source(self, file: iotools.TextFile) -> dict:
        """Load argument definitions from the source file."""
        pass

    def get(self, key: str, default: typing.Any=None, format: str=None):
        """Get (and optionally format) an argument or return `default`."""
        if not format:
            return super().get(key, default)
        raise NotImplementedError(f"Unrecognized format {format}")

    def format(self, mode: str):
        """Format this object according to `mode`."""
        return {key: self.get(key, format=mode) for key in self}


class BaseTypesH(SourceFile):
    """A representation of constant values in EPREM `baseTypes.h`."""

    def __init__(self, source: typing.Union[str, pathlib.Path]=None) -> None:
        super().__init__('_BASETYPES_H', 'baseTypes.h', source)
        self._types = None
        self._cache = {}

    def standardize(self, loaded: dict):
        return super().standardize(loaded)

    def load_from_source(self, file: iotools.TextFile) -> dict:
        typedef = BaseTypeDef()
        return file.extract(typedef.match, typedef.parse)

    def get(self, key: str, default: typing.Any=None, format: str=None):
        if format == 'json':
            value = self.definitions.get(key, default)
            if isinstance(value, algebra.Expression):
                return value.format(separator=' * ')
            return value
        return super().get(key, default, format)

    def __getitem__(self, key: str):
        """Access constants by keyword."""
        if key in self._cache:
            return self._cache[key]
        if key in self.definitions:
            value = self._compute(key)
            self._cache[key] = value
            return value
        raise KeyError(f"No {key!r} in {self!r}")

    def _compute(self, key: str) -> numbers.Real:
        """Compute the value of a defined constant."""
        target = self.definitions[key]
        realtype = self.types.get(key)
        if isinstance(target, realtype):
            return target
        if any(c in target for c in {'*', '/', 'sqrt'}):
            return self._evaluate(algebra.Expression(target))
        if realtype:
            return realtype(target)
        raise TypeError(target)

    @property
    def types(self):
        """The type of each constant."""
        if self._types is None:
            self._types = {k: v['type'] for k, v in _BASETYPES_H.items()}
        return self._types

    def _evaluate(self, expression: algebra.Expression):
        """Internal method for evaluating algebraic definitions."""
        value = 1.0
        for term in expression:
            if term.base in self.definitions:
                value *= float(term(self._compute(term.base)))
            elif term.base == '1':
                value *= float(term)
        return value


class FunctionCall:
    """Pattern parser for function calls."""

    # TODO: 
    # - __init__ should take arguments that allow the user to define possible
    #   function names (e.g., 'readInt', 'readDouble') and target-variable
    #   prefix (e.g., 'config').
    # - It may be better to just grab everything in parentheses as `args` and
    #   let `parse` sort them out.

    @property
    def pattern(self):
        return re.compile(r"""
            # the start of the string, followed by optional whitespace
            \A\s*
            # the name of the C struct
            config\.
            # the name of the attribute
            (?P<name>\w+)
            # equals sign surrounded by optional whitespace
            \s*=\s*
            # kludge for `(char*)readString(...)`
            (?:\(char\*\))?
            # the name of the file-reading method
            (?P<mode>read(?:Int|Double|DoubleArray|String))
            # beginning of function call
            \(
                # parameter name in config file
                (?P<alias>\"\w*\")
                # optional whitespace followed by a comma
                \s*\,
                # the remaining arguments
                (?P<args>.*?)
            # end of function call
            \)
            # C statement terminator
            \;
            # optional whitespace, followed by the end of the string
            \s*\Z
        """, re.VERBOSE)

    def match(self, line: str):
        """Identify lines that read config-file input."""
        if match := self.pattern.match(line.strip()):
            return match.groupdict()

    def parse(self, parsable: typing.Dict[str, str]):
        """Parse a line that reads config-file input."""
        parsed = {}
        name = parsable['name']
        if alias := parsable['alias']:
            a = alias.strip('"')
            if a != name:
                parsed['alias'] = a
        mode = parsable['mode']
        args = parsable['args'].split(',')
        parsed.update(self._normalize(mode, args))
        return name, parsed

    _keys = {
        'readInt': ['default', 'minimum', 'maximum'],
        'readDouble': ['default', 'minimum', 'maximum'],
        'readString': ['default'],
        'readDoubleArray': ['size', 'default'],
    }

    def _normalize(self, mode: str, args: typing.Iterable[str]):
        """Determine argument type and keys from `mode`."""
        pairs = dict(zip(self._keys[mode], [arg.strip(' "') for arg in args]))
        return {'mode': mode, **pairs}


class VariableDefinition:
    """Pattern parser for variable definitions."""

    # TODO:
    # - __init__ should take arguments that allow the user to define the
    #   variable type (e.g., 'Scalar_t').
    # - Can the pattern identify things like `size` and `value` with optional
    #   user input, or does `parse` need to handle that?

    @property
    def pattern(self):
        return re.compile(r"""
            # the start of the string, followed by optional whitespace
            \A\s*
            # type declaration
            Scalar\_t
            # optional whitespace
            \s*
            # variable name
            (?P<name>\w+)
            # array size
            \[(?P<size>\d+)\]
            # equals sign surrounded by optional whitespace
            \s*=\s*
            # array value(s)
            \{(?P<value>\d*(?:\.\d+)?)\}
            # C statement terminator
            \;
            # optional whitespace, followed by the end of the string
            \s*\Z
        """, re.VERBOSE)

    def match(self, line: str):
        """Identify lines that define a variable."""
        if match := self.pattern.match(line.strip()):
            return match.groupdict()

    def parse(self, parsable: typing.Dict[str, str]):
        """Parse information about a defined variable."""
        parsed = {
            'size': int(parsable['size']),
            'value': parsable['value'],
        }
        return parsable['name'], parsed


_RT = typing.TypeVar('_RT')
def soft_convert(
    string: str,
    convert: typing.Union[_RT, typing.Callable[[str], _RT]],
    acceptable: typing.Union[Exception, typing.Tuple[Exception]]=Exception,
) -> typing.Union[str, _RT]:
    """Convert a string to a different type, if possible.

    This function will use the given callable to attempt to convert the given
    string to a different type. If the conversion fails, this function will
    return the unconverted string.

    Parameters
    ----------
    string
        The string to convert.

    convert : callable
        A callable object that accepts a string argument and returns the
        converted value.

    acceptable : exception or iterable of exceptions
        One or more types of valid exception classes that this function should
        ignore when attempting to convert `string`. The default behavior is to
        accept any subclass of `Exception`. Passing arguments to this keyword
        therefore allows the caller to limit the scope of ignorable errors.

    Returns
    -------
    converted type or string
        The converted value of the given string, if conversion succeeded, or the
        unconverted string.
    """
    try:
        value = convert(string)
    except acceptable or Exception:
        return string
    return value


class ConfigurationC(SourceFile):
    """A representation of default arguments in EPREM `configuration.c`."""

    _types = iterables.Bijection(
        {
            'readInt': int,
            'readDouble': float,
            'readString': str,
            'readDoubleArray': list,
        }
    )

    def __init__(self, source: typing.Union[str, pathlib.Path]=None) -> None:
        super().__init__('_CONFIGURATION_C', 'configuration.c', source)

    def standardize(self, loaded: dict):
        return self._replace(loaded, 'mode', 'type', self._types)

    def load_from_source(self, file: iotools.TextFile) -> dict:
        assignments = self._get_assignments(file)
        arrays = self._get_array_defaults(file)
        subs = {
            key: {
                'mode': assigned['mode'],
                'default': arrays[assigned['default']]['value']
            } for key, assigned in assignments.items()
            if assigned['default'] in arrays
        }
        return {
            key: subs.get(key, attrs)
            for key, attrs in assignments.items()
        }

    def __getitem__(self, key: str):
        """Request a reference object by parameter name."""
        if key in self.definitions:
            return self.definitions[key]
        raise KeyError(f"No reference information for {key!r}")

    def get(self, key: str, default: typing.Any=None, format: str=None):
        if format == 'json':
            modes = self._types.invert()
            loaded = self._replace(self.definitions, 'type', 'mode', modes)
            return loaded.get(key, default)
        return super().get(key, default, format)

    def _get_assignments(self, file: iotools.TextFile):
        """Get the assigned default values from function calls."""
        pattern = FunctionCall()
        return file.extract(pattern.match, pattern.parse)

    def _get_array_defaults(self, file: iotools.TextFile):
        """Get all array default values defined in the source."""
        pattern = VariableDefinition()
        return file.extract(pattern.match, pattern.parse)

    def _replace(
        self,
        mapping: typing.Mapping[str, typing.Mapping],
        old: str,
        new: str,
        conversion: typing.Mapping,
    ) -> typing.Mapping[str, typing.Mapping]:
        """Convert values in the interior mappings of `mapping`."""
        return {
            key: {
                new: conversion[interior[old]],
                **{k: v for k, v in interior.items() if k != old}
            } for key, interior in mapping.items()
        }


class ConfigKeyError(KeyError):
    pass


class ConfigFile(iterables.MappingBase):
    """A class to handle EPREM run configuration files.

    Parameters
    ---------------------
    filepath : string or path
        The full path to the simulation-run config file to read. This class will
        convert to input string or `pathlib.Path` object into a fully-qualified
        read-only path.

    comments : list of strings, default='#'
        List of single-character strings to interpret as signifying a comment
        line. This method will automatically ignore lines that begin with either
        the empty string or the newline character.

    Notes
    -----
    This interface was designed to provide a faithful representation of the
    information in a given configuration file. A few notable consequences are as
    follows:
    - The look-up methods do not accept aliases as keys.
    - The parsing routine does not attempt to cast parsed values from strings to
      their underlying types, meaning all values are internally represented as
      strings. This only applies to each instance as a whole, since the object
      returned by look-up methods may perform type casting.
    - The unit associated with a specific value, when available, are those
      consistent with the convetions of EPREM (cf. configuration.c) and do not
      necessarily conform to a particular unit system. For example, many
      reference lengths or distances (e.g., reference mean free path or observer
      positions) are in au despite the fact that EPREM works with lengths in cm.
    """
    def __init__(
        self,
        filepath: typing.Union[str, pathlib.Path],
        comments: typing.List[str]=None,
    ) -> None:
        self.filepath = iotools.ReadOnlyPath(filepath)
        self.comments = comments or ['#']
        self.KeyError = ConfigKeyError
        self.parsed = self._parse()
        super().__init__(tuple(self.parsed))

    def __getitem__(self, key: str):
        """Get a value and unit for a configuration option."""
        if key in self.parsed:
            return self.parsed[key]
        raise self.KeyError(key)

    def _parse(self) -> typing.Dict[str, str]:
        """Parse an EPREM config file into a dictionary.

        This method opens the file that the given filepath points to and reads
        it line-by-line. It ignores lines that begin with a valid comment
        character and parses the rest into key-value pairs. It will
        automatically strip out inline comments.
        """
        pairs = {}
        with self.filepath.open('r') as fp:
            for line in fp:
                line = line.rstrip('\n')
                if line == '' or line == '\n' or line[0] in self.comments:
                    continue
                key, _tmp = line.split('=')
                value = iotools.strip_inline_comments(_tmp, self.comments)
                pairs[key] = value
        return pairs

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return str(self.parsed)


class Runtime(iterables.MappingBase):
    """Parameter arguments relevant to an EPREM run.

    An instance of this class represents metadata, default values, and user
    arguments for parameters available to a specific EPREM simulation run.
    """

    def __init__(
        self,
        source_path: iotools.PathLike,
        config_path: iotools.PathLike=None,
        **kwargs
    ) -> None:
        """
        Parameters
        ----------
        source_path : path-like
            The location of the directory containing relevant EPREM source code.

        config_path : path-like, default=None
            The location of the EPREM configuration file containing
            user-provided parameter values for a particular EPREM simulation
            run. May be omitted, in which case this class will use default
            argument values.

        Notes
        -----
        Path-like arguments may be any path-like object (e.g., `str`,
        `pathlib.Path`), may be relative, and may contain standard wildcard
        characters (e.g., `~`).
        """
        self._paths = {
            'source': source_path,
            'config': config_path,
        }
        self._user = None
        self._basetypes = None
        self._reference = None
        super().__init__(tuple(self.reference))
        # BUG: These `kwargs` will persist even if the user updates the
        # config-file path, which could create an inconsistency.
        self._kwargs = kwargs
        self._current = None

    def __getitem__(self, name: str):
        """Retrieve a parameter value by name."""
        if name in self.reference:
            return self._evaluate(name)
        raise KeyError(f"Unknown parameter {name!r}")

    def path_to(self, *current: str, **new: iotools.PathLike):
        """Get or set paths."""
        if current and new:
            raise TypeError("Can't simultaneously get and set paths")
        if current:
            if len(current) == 1:
                return self._paths[current[0]]
            return [self._paths[name] for name in current]
        if new:
            self._paths.update(new)
            self._reset(paths=new)
            return self
        raise TypeError("No paths to get or set")

    def _reset(self, paths: typing.Container=None):
        """Reset attributes based on updates to other attributes."""
        paths = paths or {}
        self._current = None
        if 'config' in paths:
            self._user = None
        if 'source' in paths:
            self._basetypes = None
            self._reference = None

    @property
    def basetypes(self):
        """Values of constants defined in `baseTypes.h`."""
        if self._basetypes is None:
            self._basetypes = BaseTypesH(self._paths['source'])
        return self._basetypes

    @property
    def user(self) -> typing.Mapping[str, str]:
        """The user-provided parameter values."""
        if self._user is None:
            if path := self._paths['config']:
                self._user = ConfigFile(path, **self._kwargs)
            else:
                self._user = {}
        return self._user

    @property
    def reference(self):
        """Parameter metadata from `configuration.c`."""
        if self._reference is None:
            self._reference = ConfigurationC(self._paths['source'])
        return self._reference

    _special_cases = {
        'N_PROCS': None,
        'third': 1/3,
    }

    def _evaluate(self, current: typing.Union[str, numbers.Real]):
        """Compute the final value of a current."""
        if isinstance(current, numbers.Real):
            return current
        if value := numerical.cast(current, strict=False):
            return value
        if current in self._special_cases:
            return self._special_cases[current]
        argument = self._get_argument(current)
        if argument is not None:
            definition = argument['definition']
            if isinstance(definition, argument['type']):
                return definition
            return self._evaluate(definition)
        if isinstance(current, str):
            return self._resolve(current)
        raise TypeError(f"Can't evaluate {current!r}") from None

    _struct_member = re.compile(r'\Aconfig\.\w*\Z')

    def _resolve(self, definition: str):
        """Resolve a parameter definition into simpler components."""
        if self._struct_member.match(definition):
            return self._evaluate(definition.replace('config.', ''))
        if result := self._compute_sum(definition):
            return result
        if any(c in definition for c in {'*', '/'}):
            expression = algebra.Expression(definition)
            evaluated = [
                term.coefficient * self._evaluate(term.base)**term.exponent
                for term in expression
            ]
            return functools.reduce(lambda x,y: x*y, evaluated)
        raise TypeError(f"Can't resolve {definition!r}") from None

    def _compute_sum(self, arg: str) -> numbers.Real:
        """Compute the sum of two known parameters."""
        # HACK: This is only designed to handle strings that contain a single
        # additive operator joining two arguments that `_evaluate` already knows
        # how to handle.
        for operator in ('+', '-'):
            if operator in arg:
                terms = [
                    self._evaluate(s.strip())
                    for s in arg.split(operator)
                ]
                return terms[0] + float(f'{operator}1')*terms[1]

    def _get_argument(self, name: str):
        """Get the current definition and type of a parameter."""
        if self._current is None:
            self._current = {
                key: self._convert(
                    self.user.get(key) or parameter['default'],
                    parameter['type'],
                )
                for key, parameter in self.reference.items()
            }
        if name in self._current:
            return {
                'definition': self._current[name],
                'type': self.reference[name]['type']
            }
        if name in self.basetypes:
            return {
                'definition': self.basetypes[name],
                'type': type(self.basetypes[name]),
            }

    def _convert(self, arg: typing.Union[str, _RT], realtype: _RT):
        """Convert `arg` to its real type, if necessary and possible."""
        if isinstance(arg, realtype):
            return arg
        if realtype in {int, float}:
            return soft_convert(arg, realtype)
        if realtype == list:
            return soft_convert(arg, iterables.string_to_list)
        return arg


class Arguments(aliased.Mapping):
    """Aliased access to EPREM parameter arguments."""

    def __init__(
        self,
        runtime: Runtime=None,
        **runtime_init,
    ) -> None:
        """Initialize an instance of this class.

        The caller may pass an existing instance of `~parameters.Runtime` or the
        arguments necessary to create the appropriate instance. See
        documentation at `~parameters.Runtime` for information. Note that this
        class will use an existing instance of `~parameters.Runtime` if
        available, without checking for the presence of additional arguments.

        Parameters
        ----------
        runtime : optional
            An instance of `parameters.Runtime` initialized with paths relevant
            to the simulation run under analysis.

        **runtime_init : optional
            See `~parameters.Runtime`.
        """
        runtime = runtime or Runtime(**runtime_init)
        super().__init__(mapping=self._build_mapping(runtime))

    def __getitem__(self, key: str):
        """Get the value (and unit, if applicable) of a named parameter."""
        try:
            parameter = super().__getitem__(key)
        except KeyError:
            raise KeyError(f"No parameter corresponding to '{key}'") from None
        value = parameter['value']
        if unit := parameter['unit']:
            if isinstance(value, list):
                return [quantities.Scalar(v, unit) for v in value]
            return quantities.Scalar(value, unit)
        return value

    def _build_mapping(self, runtime: Runtime):
        """Build the mapping of available parameters."""
        values = {
            **{
                key: info.get('default')
                for key, info in _LOCAL.items()
            },
            **dict(runtime),
        }
        keys = tuple(set(tuple(_LOCAL) + tuple(runtime)))
        base = {
            tuple([key, *_ALIASES.get(key, [])]): {
                'unit': _UNITS.get(key),
                'value': values.get(key),
            }
            for key in keys
        }
        return aliased.MutableMapping(base)


_LOCAL = {
    'minimum_energy': {
        'aliases': ['Emin', 'minimum energy'],
        'unit': 'MeV',
        'default': 0.0,
    },
    'reference energy': {
        'aliases': ['energy0'],
        'unit': 'MeV',
        'default': 1.0,
    },
    'reference radius': {
        'aliases': ['r0'],
        'unit': 'au',
        'default': 1.0,
    },
}
"""Metadata for post-processing parameters."""


_CONFIGURATION_C = {
    'boundaryFunctAmplitude': {
        'aliases': ['J0'],
        'unit': '1 / (cm^2 * s * sr * (MeV/nuc))',
    },
    'boundaryFunctXi': {
        'aliases': ['xi'],
        'unit': '1',
    },
    'boundaryFunctBeta': {
        'aliases': ['beta'],
        'unit': '1',
    },
    'boundaryFunctGamma': {
        'aliases': ['gamma'],
        'unit': '1',
    },
    'boundaryFunctEcutoff': {
        'aliases': ['E0'],
        'unit': 'MeV',
    },
    'kperxkpar': {
        'aliases': ['kper_kpar', 'kper/kpar', 'kper / kpar'],
        'unit': '1',
    },
    'lamo': {
        'aliases': ['lam0', 'lambda0'],
        'unit': 'au',
    },
    'mfpRadialPower': {
        'aliases': ['mfp_radial_power'],
        'unit': '1',
    },
    'rigidityPower': {
        'aliases': ['rigidity_power'],
        'unit': '1',
    },
    'flowMag': {
        'aliases': [],
        'unit': 'km/s',
    },
    'mhdDensityAu': {
        'aliases': [],
        'unit': 'cm^-3',
    },
    'mhdBAu': {
        'aliases': [],
        'unit': 'G',
    },
    'omegaSun': {
        'aliases': [],
        'unit': 'rad * cm / (au * s)',
    },
    'mass': {
        'aliases': [],
        'unit': 'nucleon',
    },
    'charge': {
        'aliases': [],
        'unit': 'e',
    },
    'minInjectionEnergy': {
        'aliases': [],
        'unit': 'MeV',
    },
    'maxInjectionEnergy': {
        'aliases': [],
        'unit': 'MeV',
    },
    'shockInjectionFactor': {
        'aliases': [],
        'unit': '1',
    },
    'shockDetectPercent': {
        'aliases': [],
        'unit': '1',
    },
    'rScale': {
        'aliases': [],
        'unit': 'au',
    },
    'simStartTime': {
        'aliases': [],
        'unit': 'day',
    },
    'simStopTime': {
        'aliases': [],
        'unit': 'day',
    },
    'eMin': {
        'aliases': [],
        'unit': 'MeV/nuc',
    },
    'eMax': {
        'aliases': [],
        'unit': 'MeV/nuc',
    },
    'numObservers': {
        'aliases': [],
        'unit': '1',
    },
    'obsR': {
        'aliases': [],
        'unit': 'au',
    },
    'obsTheta': {
        'aliases': [],
        'unit': 'rad',
    },
    'obsPhi': {
        'aliases': [],
        'unit': 'rad',
    },
    'idw_p': {
        'aliases': [],
        'unit': '1',
    },
    'idealShockSharpness': {
        'aliases': [],
        'unit': '1',
    },
    'idealShockScaleLength': {
        'aliases': [],
        'unit': 'au',
    },
    'idealShockJump': {
        'aliases': [],
        'unit': '1',
    },
    'idealShockSpeed': {
        'aliases': [],
        'unit': 'km/s',
    },
    'idealShockInitTime': {
        'aliases': [],
        'unit': 'day',
    },
    'idealShockTheta': {
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockPhi': {
        'aliases': [],
        'unit': 'rad',
    },
    'idealShockWidth': {
        'aliases': [],
        'unit': 'rad',
    },
    'tDel': {
        'aliases': [],
        'unit': 'day',
    },
    'gammaElow': {
        'aliases': [],
        'unit': '1',
    },
    'gammaEhigh': {
        'aliases': [],
        'unit': '1',
    },
    'masInitTimeStep': {
        'aliases': [],
        'unit': 'day',
    },
    'masStartTime': {
        'aliases': [],
        'unit': 'day',
    },
    'epEquilibriumCalcDuration': {
        'aliases': [],
        'unit': 'day',
    },
    'preEruptionDuration': {
        'aliases': [],
        'unit': 'day',
    },
    'pointObserverOutput': {
        'aliases': [],
        'unit': None,
    },
    'enlilCouple': {
        'aliases': [],
        'unit': None,
    },
    'outputFloat': {
        'aliases': [],
        'unit': None,
    },
    'numRowsPerFace': {
        'aliases': [],
        'unit': None,
    },
    'numColumnsPerFace': {
        'aliases': [],
        'unit': None,
    },
    'numNodesPerStream': {
        'aliases': [],
        'unit': None,
    },
    'numEnergySteps': {
        'aliases': [],
        'unit': None,
    },
    'numMuSteps': {
        'aliases': [],
        'unit': None,
    },
    'useDrift': {
        'aliases': [],
        'unit': None,
    },
    'useShellDiffusion': {
        'aliases': [],
        'unit': None,
    },
    'unifiedOutput': {
        'aliases': [],
        'unit': None,
    },
    'streamFluxOutput': {
        'aliases': [],
        'unit': None,
    },
    'epremDomain': {
        'aliases': [],
        'unit': None,
    },
    'dumpFreq': {
        'aliases': [],
        'unit': None,
    },
    'idealShock': {
        'aliases': [],
        'unit': None,
    },
    'shockSolver': {
        'aliases': [],
        'unit': None,
    },
    'fluxLimiter': {
        'aliases': [],
        'unit': None,
    },
    'numEpSteps': {
        'aliases': [],
        'unit': None,
    },
    'useParallelDiffusion': {
        'aliases': [],
        'unit': None,
    },
    'useAdiabaticChange': {
        'aliases': [],
        'unit': None,
    },
    'useAdiabaticFocus': {
        'aliases': [],
        'unit': None,
    },
    'numSpecies': {
        'aliases': [],
        'unit': None,
    },
    'boundaryFunctionInitDomain': {
        'aliases': [],
        'unit': None,
    },
    'checkSeedPopulation': {
        'aliases': [],
        'unit': None,
    },
    'subTimeCouple': {
        'aliases': [],
        'unit': None,
    },
    'FailModeDump': {
        'aliases': [],
        'unit': None,
    },
    'masCouple': {
        'aliases': [],
        'unit': None,
    },
    'masDirectory': {
        'aliases': [],
        'unit': None,
    },
    'masInitFromOuterBoundary': {
        'aliases': [],
        'unit': None,
    },
    'masRotateSolution': {
        'aliases': [],
        'unit': None,
    },
    'useMasSteadyStateDt': {
        'aliases': [],
        'unit': None,
    },
    'masDigits': {
        'aliases': [],
        'unit': None,
    },
}
"""Metadata for parameters defined in `configuration.c`."""

_BASETYPES_H = {
    'T': {
        'info': 'True',
        'unit': None,
        'type': int,
    },
    'F': {
        'info': 'False',
        'unit': None,
        'type': int,
    },
    'PI': {
        'info': 'The value of π.',
        'unit': None,
        'type': float,
    },
    'TWO_PI': {
        'info': 'The value of 2π.',
        'unit': None,
        'type': float,
    },
    'VERYSMALL': {
        'info': 'A very small value.',
        'unit': None,
        'type': float,
    },
    'BADVALUE': {
        'info': 'A bad (invalid) float value.',
        'unit': None,
        'type': float,
    },
    'BADINT': {
        'info': 'A bad (invalid) integer value.',
        'unit': None,
        'type': int,
    },
    'MP': {
        'info': 'The proton mass.',
        'unit': 'g',
        'type': float,
    },
    'EV': {
        'info': 'The conversion from eVs to ergs.',
        'unit': 'erg/eV',
        'type': float,
    },
    'MEV': {
        'info': 'The conversion from MeVs to ergs.',
        'unit': 'erg/MeV',
        'type': float,
    },
    'GEV': {
        'info': 'The conversion from GeVs to ergs.',
        'unit': 'erg/GeV',
        'type': float,
    },
    'Q': {
        'info': 'The proton charge.',
        'unit': 'statC',
        'type': float,
    },
    'C': {
        'info': 'The speed of light.',
        'unit': 'cm/s',
        'type': float,
    },
    'MZERO': {
        'info': 'The proton rest-mass energy in GeV.',
        'unit': 'GeV',
        'type': float,
    },
    'AU': {
        'info': 'One astronomical unit.',
        'unit': 'cm',
        'type': float,
    },
    'RSUN': {
        'info': 'The value of the solar radius.',
        'unit': 'cm',
        'type': float,
    },
    'RSAU': {
        'info': 'The number of solar radii per au.',
        'unit': '1',
        'type': float,
    },
    'TAU': {
        'info': 'The canonical EPREM time scale.',
        'unit': 's',
        'type': float,
    },
    'DAY': {
        'info': 'The conversion from EPREM time steps to Julian days.',
        'unit': 'day',
        'type': float,
    },
    'MHD_DENSITY_NORM': {
        'info': 'The normalization factor for density.',
        'unit': '1',
        'type': float,
    },
    'MHD_B_NORM': {
        'info': 'The normalization for magnetic fields.',
        'unit': '1',
        'type': float,
    },
    'OM': {
        'info': 'The normalization for ion gyrofrequency.',
        'unit': '1',
        'type': float,
    },
    'FCONVERT': {
        'info': 'The conversion from distribution to flux.',
        'unit': '1',
        'type': float,
    },
    'VOLT': {
        'info': 'The conversion from volts to statvolts.',
        'unit': '1',
        'type': float,
    },
    'THRESH': {
        'info': 'The threshold for perpendicular diffusion.',
        'unit': '1',
        'type': float,
    },
    'MAS_TIME_NORM': {
        'info': 'The MAS time normalization factor.',
        'unit': '1',
        'type': float,
    },
    'MAS_LENGTH_NORM': {
        'info': 'The MAS length normalization factor.',
        'unit': '1',
        'type': float,
    },
    'MAS_RHO_NORM': {
        'info': 'The MAS plasma-density normalization.',
        'unit': '1',
        'type': float,
    },
    'MAS_TIME_CONVERT': {
        'info': 'The time conversion from MAS units.',
        'unit': '1',
        'type': float,
    },
    'MAS_V_CONVERT': {
        'info': 'The velocity conversion from MAS units.',
        'unit': '1',
        'type': float,
    },
    'MAS_RHO_CONVERT': {
        'info': 'The density conversion from MAS units.',
        'unit': '1',
        'type': float,
    },
    'MAS_B_CONVERT': {
        'info': 'The magnetic field conversion from MAS units.',
        'unit': '1',
        'type': float,
    },
    'MAX_STRING_SIZE': {
        'info': '',
        'unit': None,
        'type': int,
    },
    'MHD_DEFAULT': {
        'info': 'Use the default MHD solver.',
        'unit': None,
        'type': int,
    },
    'MHD_ENLIL': {
        'info': 'Use ENLIL for MHD values.',
        'unit': None,
        'type': int,
    },
    'MHD_LFMH': {
        'info': 'Use LFM for MHD values.',
        'unit': None,
        'type': int,
    },
    'MHD_BATSRUS': {
        'info': 'Use BATS-R-US for MHD values.',
        'unit': None,
        'type': int,
    },
    'MHD_MAS': {
        'info': 'Use MAS for MHD values.',
        'unit': None,
        'type': int,
    },
    'NUM_MPI_BOUNDARY_FLDS': {
        'info': 'Number of MPI psuedo-fields for use in creating MPI typedefs.',
        'unit': None,
        'type': int,
    },
}
"""Metadata for parameters defined in `baseTypes.h`."""

_metadata = {**_LOCAL, **_CONFIGURATION_C, **_BASETYPES_H}
"""Combined metadata dictionary for internal use."""

_ALIASES = {
    key: info.get('aliases')
    for key, info in _metadata.items()
}
"""Collection of aliases from metadata."""

_UNITS = {
    key: info.get('unit')
    for key, info in _metadata.items()
}
"""Collection of units from metadata."""


DIRECTORY = pathlib.Path(__file__).expanduser().resolve().parent
"""The full directory containing this module."""


def generate_defaults(path: iotools.PathLike):
    """Generate default arguments from the EPREM source code in `path`."""
    obj = {
        '_BASETYPES_H': {**BaseTypesH(path).format('json')},
        '_CONFIGURATION_C': {**ConfigurationC(path).format('json')},
    }
    outpath = pathlib.Path(__file__).with_suffix('.json')
    with outpath.open('w') as fp:
        json.dump(obj, fp, indent=4, sort_keys=True)


if __name__ == '__main__':
    doclines = __doc__.split('\n')
    parser = argparse.ArgumentParser(
        description=doclines[0],
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        '-r',
        '--generate_defaults',
        help=generate_defaults.__doc__.replace('`path`', 'SRC'),
        metavar='SRC',
    )
    args = parser.parse_args()
    kwargs = vars(args)
    if 'generate_defaults' in kwargs:
        generate_defaults(kwargs['generate_defaults'])




import typing
import inspect
import sys

import numpy
from scipy import integrate

from goats.core import aliased
from goats.core import datasets
from goats.core import iterables
from goats.core import numerical
from goats.core import observables
from goats.core import physical


registry = iterables.ObjectRegistry(object_key='method')


MKS = physical.Constants('mks')


@registry.register
def x(
    r: numpy.ndarray,
    theta: numpy.ndarray,
    phi: numpy.ndarray,
) -> numpy.ndarray:
    """The x-axis component given by x = r·sin(θ)cos(φ)"""
    return r * numpy.sin(theta) * numpy.cos(phi)


@registry.register
def y(
    r: numpy.ndarray,
    theta: numpy.ndarray,
    phi: numpy.ndarray,
) -> numpy.ndarray:
    """The y-axis component given by y = r·sin(θ)sin(φ)"""
    return r * numpy.sin(theta) * numpy.sin(phi)


@registry.register
def z(
    r: numpy.ndarray,
    theta: numpy.ndarray,
) -> numpy.ndarray:
    """The z-axis component given by z = r·cos(θ)"""
    return r * numpy.cos(theta)


@registry.register
def b_mag(
    br: numpy.ndarray,
    bt: numpy.ndarray,
    bp: numpy.ndarray,
) -> numpy.ndarray:
    """The magnetic-field magnitude."""
    return numpy.sqrt(br**2 + bt**2 + bp**2)


@registry.register
def v_mag(
    vr: numpy.ndarray,
    vt: numpy.ndarray,
    vp: numpy.ndarray,
) -> numpy.ndarray:
    """The velocity-field magnitude."""
    return numpy.sqrt(vr**2 + vt**2 + vp**2)


@registry.register
def bv_mag(
    br: numpy.ndarray,
    bt: numpy.ndarray,
    bp: numpy.ndarray,
    vr: numpy.ndarray,
    vt: numpy.ndarray,
    vp: numpy.ndarray,
) -> numpy.ndarray:
    """|B·V|"""
    return b_mag(br, bt, bp) * v_mag(vr, vt, vp)


@registry.register
def flow_angle(
    br: numpy.ndarray,
    bt: numpy.ndarray,
    bp: numpy.ndarray,
    vr: numpy.ndarray,
    vt: numpy.ndarray,
    vp: numpy.ndarray,
) -> numpy.ndarray:
    """The angle between the magnetic- and velocity-field vectors."""
    b_dot_v = v_para(br, bt, bp, vr, vt, vp)
    bv = bv_mag(br, bt, bp, vr, vt, vp)
    arg = numpy.array(b_dot_v / bv)
    arg[arg < -1.0] = -1.0
    arg[arg > +1.0] = +1.0
    return numpy.arccos(arg)


@registry.register
def v_para(
    vr: numpy.ndarray,
    vt: numpy.ndarray,
    vp: numpy.ndarray,
    br: numpy.ndarray,
    bt: numpy.ndarray,
    bp: numpy.ndarray,
) -> numpy.ndarray:
    """The velocity-field component parallel to the magnetic field."""
    return (br*vr + bt*vt + bp*vp) / b_mag(br, bt, bp)


@registry.register
def v_perp(
    vr: numpy.ndarray,
    vt: numpy.ndarray,
    vp: numpy.ndarray,
    br: numpy.ndarray,
    bt: numpy.ndarray,
    bp: numpy.ndarray,
) -> numpy.ndarray:
    """The velocity-field component parallel to the magnetic field."""
    v_mag = v_mag(vr, vt, vp)
    v_para = v_para(vr, vt, vp, br, bt, bp)
    return numpy.sqrt(v_mag**2 - v_para**2)


@registry.register
def div_v(
    rho: numpy.ndarray,
    time: numpy.ndarray,
) -> numpy.ndarray:
    """Divergence of the velocity field.

    The method computes ∇·V from the continuity equation. The continuity
    equation in the co-moving frame can be written as

        1  dn
        --- -- = - div(V)
        n  dt

    where n represents density and dn/dt is the convective derivative. This
    method rewrites the time derivative as a logarithmic derivative and
    computes

                    d(ln n)
        div(V) = - --------
                    dt
    """
    return -1.0*numpy.gradient(numpy.log(rho), time, axis=0)


@registry.register
def density_ratio(
    rho: numpy.ndarray,
) -> numpy.ndarray:
    """Compute the ratio of the local density to the initial density."""
    return rho / rho[0, ...]


@registry.register
def rigidity(
    energy: typing.Union[float, typing.Iterable[float]],
    mass: typing.Union[numpy.ndarray, float],
    charge: typing.Union[numpy.ndarray, float],
) -> numpy.ndarray:
    """The magnetic rigidity.

    This function assumes that `energy` represents the kinetic energy of a
    particle. It then computes the particle momentum based on the
    relativistic relation E² = (`energy` + mc²)² = p²c² + (mc²)². Finally,
    it divides the momentum by the particle charge to get rigidity.
    """
    c = float(MKS['c'])
    total_energy = numpy.array(energy, ndmin=1) + mass*c**2
    rest_energy = mass*c**2
    p = numpy.sqrt(total_energy**2 - rest_energy**2) / c
    return p / charge


@registry.register
def mean_free_path(
    r: numpy.ndarray,
    energy: numpy.ndarray,
    mass: typing.Union[numpy.ndarray, float],
    charge: typing.Union[numpy.ndarray, float],
    mfp_radial_power: float,
    rigidity_power: float,
    lambda0: float,
) -> numpy.ndarray:
    """The scattering mean free path."""
    rg = rigidity(energy, mass, charge)
    rg0 = rigidity(1e6 * float(MKS['eV']), mass, charge)
    mfp = numpy.tensordot(
        pow(r / float(MKS['au']), float(mfp_radial_power)),
        pow(rg / rg0, float(rigidity_power)),
        axes=0,
    )
    return mfp * lambda0


@registry.register
def acceleration_rate(
    vr: numpy.ndarray,
    vtheta: numpy.ndarray,
    vphi: numpy.ndarray,
    vparticle: numpy.ndarray,
    angle: numpy.ndarray,
    mfp: numpy.ndarray,
    kper_kpar: float,
) -> numpy.ndarray:
    """Compute the acceleration rate for this dataset."""
    v1_dot_v2 = (
        _ahead(vr)*_behind(vr)
        + _ahead(vtheta)*_behind(vtheta)
        + _ahead(vphi)*_behind(vphi)
    )
    v = v_mag(vr, vtheta, vphi)
    v1 = _ahead(v)
    delta_v = v1 * numpy.abs(1 - v1_dot_v2 / (v1*v1))
    kxx_kpar = numpy.cos(angle)**2 + kper_kpar*numpy.sin(angle)**2
    vpara = numpy.tensordot(kxx_kpar, vparticle, axes=0)
    kappa = mfp * (vpara / 3.0)
    forward = inverse = (2, 3, 0, 1)
    tmp = kappa.transpose(forward) / v
    kxx_vx = tmp.transpose(inverse)
    delta_x = 3.0 * (_ahead(kxx_vx) + _behind(kxx_vx))
    tmp = delta_v / delta_x.transpose(forward)
    return tmp.transpose(inverse)

def _ahead(arr: numpy.ndarray) -> numpy.ndarray:
    """Extract array values at the shell immediately ahead."""
    nshells = arr.shape[1]
    idx = [*list(range(1, nshells)), nshells-1]
    return arr[:, idx, ...]

def _behind(arr: numpy.ndarray) -> numpy.ndarray:
    """Extract array values at the shell immediately behind."""
    nshells = arr.shape[1]
    idx = [0, *list(range(nshells-1))]
    return arr[:, idx, ...]


@registry.register
def energy_density(
    speed: numpy.ndarray,
    isodist: numpy.ndarray,
) -> numpy.ndarray:
    """Compute the distribution energy density for this dataset.

    This function converts the result to energy density in ergs / cm³ before
    returning. It first approximates the integral of v⁴f(v) with respect to
    v, where v is the particle velocity and f(v) is the isotropic
    distribution. The particle velocities have units of cm/s and the
    isotropic distribution has units of s³/cm⁶, giving the intermediate
    value units of (cm/s)⁵s³/cm⁶, or 1/cm/s². Multiplying by the proton mass
    in grams yields an expression with units of g/cm/s², or erg/cm³.
    """
    v = numpy.array(speed, ndmin=1)
    dv = numpy.gradient(v, axis=-1)
    return float(MKS['mp']) * numpy.sum(v**4 * isodist * dv, axis=-1)


@registry.register
def average_energy(
    speed: numpy.ndarray,
    isodist: numpy.ndarray,
) -> numpy.ndarray:
    """Compute the average distribution energy for this dataset.

    This function converts the result to average energy in ergs before
    returning. It first approximates the integral of v²f(v) with respect to
    v, where v is the particle velocity and f(v) is the isotropic
    distribution. The particle velocities have units of cm/s and the
    isotropic distribution has units of s³/cm⁶, giving the intermediate
    value units of 1/cm³. This function then computes the energy density via
    `energy_density()`, divides that result by the intermediate result
    computed here, and further divides by 2 to represent the kinetic energy.
    The final result has units of ergs.
    """
    v = numpy.array(speed, ndmin=1)
    dv = numpy.gradient(v, axis=-1)
    normalizer = numpy.sum(v**2 * isodist * dv, axis=-1)
    epsilon = energy_density(v, isodist)
    return 0.5 * epsilon / normalizer


@registry.register(removed='pitch-angle')
def isotropic_distribution(
    dist: numpy.ndarray,
) -> numpy.ndarray:
    """Compute the isotropic distribution.

    This function computes the isotropic form of a given particle
    distribution by averaging over all pitch angles. It returns an array
    with the final dimension removed.
    """
    return numpy.mean(dist, axis=-1)


@registry.register
def flux(
    energy: typing.Union[float, typing.Iterable[float]],
    isodist: numpy.ndarray,
) -> numpy.ndarray:
    """Compute the differential energy flux of a distribution."""
    dist_to_flux = 2 * numpy.array(energy, ndmin=1) / float(MKS['mp'])**2
    return dist_to_flux * isodist


@registry.register(removed='time')
def fluence(
    flux: numpy.ndarray,
    time: numpy.ndarray,
) -> numpy.ndarray:
    """Compute the distribution fluence."""
    return integrate.simps(flux, time, axis=0)


def alt_intflux(
    energies: numpy.ndarray,
    flux: numpy.ndarray,
    minimum_energy: float,
) -> numpy.ndarray:
    """Compute the integral flux by the old algorithm.
    
    Currently exists only for testing.
    """
    x0 = energies[0, :-1]
    x1 = energies[0, 1:]
    y0 = numpy.squeeze(flux[..., :-1])
    y1 = numpy.squeeze(flux[..., 1:])
    dlnx = numpy.log10(x1) - numpy.log10(x0)
    kernel = 0.5 * dlnx * (x0*y0 + x1*y1)
    kernel[..., x1 <= minimum_energy] = 0.0
    return numpy.sum(
        kernel,
        axis=-1,
        dtype=numpy.float64,
    )


@registry.register(removed='energy')
def integral_flux(
    energies: numpy.ndarray,
    flux: numpy.ndarray,
    *minimum_energy: float,
) -> numpy.ndarray:
    """Compute the integral flux of a distribution above a given energy."""
    if energies.ndim == 1:
        energies = energies[None, :]
        if flux.ndim == 3:
            # This is under the energies.ndim check because I assume that if
            # the species dimension was squeezed from energies, it might
            # have also been squeezed from flux. In other words, I am
            # assuming that the only reason flux would have 3 (instead of 4)
            # dimensions is if the species dimensions was squeezed. Is this
            # a good assumption?
            flux = flux[:, :, None, :]
    int_flux = numpy.zeros((*flux.shape[:3], len(minimum_energy)))
    for i, bound in enumerate(minimum_energy):
        m = max(bound, sys.float_info.min)
        use_all = m < numpy.min(energies)
        for s, species_energy in enumerate(energies):
            f = flux[..., s, :]
            e = species_energy
            y, x = (f, e) if use_all else _interpolate(f, e, m)
            int_flux[..., s, i] = integrate.simps(y, x)
    return int_flux

def _interpolate(
    f: numpy.ndarray,
    x: numpy.ndarray,
    m: float,
) -> numpy.ndarray:
    """Interpolate `f` at `x = m` if necessary.

    This function uses an algorithm that interpolates `f` at `x = m` via a
    power law. If `m` is already in `x`, the interpolation leaves `f`
    unchanged. It was designed to support computations of integral flux and
    therefore currently only interpolates to a lower bound.

    This function makes copies of `x` and `f` before applying the algorithm.
    """
    xc = x.copy()
    fc = f.copy()
    i0, x0 = numerical.find_nearest(xc, m, bound='upper')
    # TODO: Make sure i0+1 is not out of range.
    beta = numpy.log(fc[..., i0+1] / fc[..., i0]) / numpy.log(xc[i0+1] / x0)
    base = numpy.full_like(beta, m / x0)
    xc[i0] = m
    fc[..., i0] *= numpy.power(base, beta)
    return fc[..., i0:], xc[i0:]


class Method(iterables.ReprStrMixin):
    """A method that acts on EPREM variables."""

    def __init__(
        self,
        __callable: typing.Callable[..., numpy.ndarray],
        __metadata: typing.Mapping[str, typing.Any]=None,
    ) -> None:
        self.callable = __callable
        self.metadata = __metadata or {}
        self.signature = inspect.signature(self.callable)
        self.parameters = tuple(self.signature.parameters)

    def __call__(self, *args, **kwargs):
        """Produce the results of this method."""
        return self.callable(*args, **kwargs)

    def __str__(self) -> str:
        """A simplified representation of this object."""
        prms = ', '.join(str(p) for p in self.parameters)
        return f"{self.callable.__qualname__}({prms})"


_METHODS = {
    observables.ALIASES[name]: Method(
        registered['method'],
        {k: v for k, v in registered.items() if k != 'method'},
    )
    for name, registered in registry.items()
}
METHODS = aliased.Mapping(_METHODS)





import typing

import numpy as np

from goats.core import base
from goats.core import aliased
from goats.core import quantities
from goats.core import algebra
from goats.core import iterables
from goats.core import physical
from goats.core import datasets
from goats.core import datatypes
from goats.eprem import functions
from goats.eprem import parameters
from goats.eprem import interpolation


MKS = quantities.MetricSystem('mks')


Observable = typing.TypeVar(
    'Observable',
    datatypes.Variable,
    functions.Function,
)
Observable = typing.Union[
    datatypes.Variable,
    functions.Function,
]


Assumption = typing.TypeVar(
    'Assumption',
    quantities.Scalar,
    typing.Iterable[quantities.Scalar],
)
Assumption = typing.Union[
    quantities.Scalar,
    typing.Iterable[quantities.Scalar],
]


Implementation = typing.TypeVar(
    'Implementation',
    Observable,
    algebra.Expression,
)
Implementation = typing.Union[
    Observable,
    algebra.Expression,
]


Dependency = typing.TypeVar(
    'Dependency',
    Observable,
    quantities.Scalar,
)
Dependency = typing.Union[
    Observable,
    quantities.Scalar,
]


Reference = typing.TypeVar(
    'Reference',
    datatypes.Variable,
    typing.Iterable,
)
Reference = typing.Union[
    datatypes.Variable,
    typing.Iterable,
]


class Compound(typing.NamedTuple):
    """An algebraic combination of primary or derived observables."""

    expression: algebra.Expression
    axes: typing.Tuple[str]
    name: str=None


class Application:
    """The object that handles evaluating all observables."""

    def __init__(
        self,
        indices: typing.Mapping[str, datatypes.Indices],
        assumptions: typing.Mapping[str, Assumption],
        observables: typing.Mapping[str, Observable],
        reference: typing.Mapping[str, Reference],
    ) -> None:
        self.indices = indices
        self.assumptions = assumptions
        self.observables = observables
        self.reference = reference

    def evaluate(self, implementation: Implementation):
        """Create a variable from the given implementation."""
        if isinstance(implementation, datatypes.Variable):
            return self._evaluate_variable(implementation)
        if isinstance(implementation, functions.Function):
            return self._evaluate_function(implementation)
        if isinstance(implementation, Compound):
            return self._evaluate_compound(implementation)
        raise TypeError(f"Unknown implementation: {type(implementation)}")

    def _evaluate_variable(self, variable: datatypes.Variable):
        """Apply relevant updates (e.g., indices) to this variable."""
        target_axes = [
            axis for axis in variable.axes if self._need_interp(axis)
        ]
        if 'radius' in self.assumptions:
            target_axes.append('shell')
        if self._is_reference(variable) or not target_axes:
            return self._standard(variable)
        return self._interpolated(variable, target_axes)

    def _standard(self, variable: datatypes.Variable):
        """Produce a new variable by subscripting the given variable."""
        indices = tuple(self.indices[axis] for axis in variable.axes)
        return variable[indices]

    def _interpolated(
        self,
        original: datatypes.Variable,
        axes: typing.Iterable[str],
    ) -> datatypes.Variable:
        """Produce a new variable by interpolating the given variable."""
        indexable = list(set(original.axes) - set(axes))
        variable = self._interpolate(original, axes)
        indices = tuple(
            self.indices[axis]
            if axis in indexable else slice(None)
            for axis in variable.axes
        )
        return variable[indices]

    def _is_reference(self, variable: datatypes.Variable):
        """True if this is an axis reference variable.

        This method attempts to determine if the given variable is one of the
        axis or axis-like reference objects used for interpolation, so calling
        code can avoid accidentally interpolating over it. NB: Two variables
        with different units will always compare false by triggering a
        `quantities.ComparisonError`.
        """
        return any(alias in self.reference for alias in variable.name)

    def _need_interp(self, axis: str):
        """True if we need to interpolate over this axis."""
        index = self.indices[axis]
        if not isinstance(index, datatypes.Coordinates):
            return False
        reference = self.reference[axis]
        targets = np.array(index.values)
        refarr = np.array(reference)
        available = [self._in_array(target, refarr) for target in targets]
        return not np.all(available)

    def _in_array(self, target: float, array: np.ndarray):
        """True if the target value is in the given array."""
        if target < np.min(array) or target > np.max(array):
            return False
        return np.any([np.isclose(target, array, atol=0.0)])

    def _interpolate(
        self,
        variable: datatypes.Variable,
        axes: typing.Iterable[str],
    ) -> datatypes.Variable:
        """Interpolate the variable over certain axes."""
        array = None
        coordinates = {
            axis: {
                'targets': np.array(indices.values),
                'reference': self.reference[axis],
            }
            for axis, indices in self.indices.items()
            if axis in axes and isinstance(indices, datatypes.Coordinates)
        }
        if 'radius' in self.assumptions:
            radii = iterables.whole(self.assumptions['radius'])
            coordinates.update(
                radius={
                    'targets': np.array([float(radius) for radius in radii]),
                    'reference': self.reference['radius'],
                }
            )
        for coordinate, current in coordinates.items():
            array = self._interpolate_coordinate(
                variable,
                current['targets'],
                current['reference'],
                coordinate=coordinate,
                workspace=array,
            )
        return datatypes.Variable(
            array,
            variable.unit,
            variable.axes,
            name=variable.name,
        )

    def _interpolate_coordinate(
        self,
        variable: datatypes.Variable,
        targets: np.ndarray,
        reference: datatypes.Variable,
        coordinate: str=None,
        workspace: np.ndarray=None,
    ) -> np.ndarray:
        """Interpolate a variable array based on a known coordinate."""
        array = np.array(variable) if workspace is None else workspace
        indices = (variable.axes.index(d) for d in reference.axes)
        dst, src = zip(*enumerate(indices))
        reordered = np.moveaxis(array, src, dst)
        interpolated = interpolation.apply(
            reordered,
            np.array(reference),
            targets,
            coordinate=coordinate,
        )
        return np.moveaxis(interpolated, dst, src)

    def _evaluate_function(self, function: functions.Function):
        """Gather dependencies and call this function."""
        deps = {p: self._get_observable(p) for p in function.parameters}
        unit = MKS.get_unit(quantity=function.quantity)
        return function(deps, unit)

    def _evaluate_compound(self, implementation: Compound):
        """Combine variables and functions based on this expression."""
        expression = implementation.expression
        variables = [
            self._get_observable(term.base) for term in expression
        ]
        exponents = [term.exponent for term in expression]
        result = variables[0] ** exponents[0]
        for variable, exponent in zip(variables[1:], exponents[1:]):
            result *= variable ** exponent
        return result

    def _get_observable(self, key: str):
        """Get an observable dependency by keyword."""
        if key in self.assumptions:
            return self.assumptions[key]
        if key in self.observables:
            observable = self.observables[key]
            if isinstance(observable, datatypes.Variable):
                return self._evaluate_variable(observable)
            if isinstance(observable, functions.Function):
                return self._evaluate_function(observable)
        raise KeyError(f"Can't find observable {key!r}")


class Interface(base.Interface):
    """A concrete EPREM observing interface."""

    def __init__(
        self,
        implementation: Implementation,
        dataset: datasets.Dataset,
        dependencies: typing.Mapping[str, Dependency]=None,
    ) -> None:
        self.implementation = implementation
        self.axes = dataset.axes
        self.dependencies = aliased.Mapping(dependencies or {})
        self._result = None
        self._context = None
        self.indices = aliased.MutableMapping.fromkeys(
            self.axes,
            value=(),
        )
        self.assumptions = aliased.MutableMapping(
            {
                k: v for k, v in self.dependencies.items(aliased=True)
                if isinstance(v, quantities.Scalar)
            }
        )
        self.observables = aliased.MutableMapping(
            {
                k: v for k, v in self.dependencies.items(aliased=True)
                if isinstance(v, (datatypes.Variable, functions.Function))
            }
        )
        axes_ref = {k: v.reference for k, v in self.axes.items(aliased=True)}
        variables = dataset.variables
        rtp_ref = {
            (k, *variables.alias(k, include=True)): variables[k]
            for k in {'radius', 'theta', 'phi'}
        }
        self.reference = aliased.Mapping({**axes_ref, **rtp_ref})

    def update_indices(self, constraints: typing.Mapping):
        """Update the instance indices based on user constraints."""
        current = self.indices.copy()
        new = {k: v for k, v in constraints.items() if k in self.axes}
        current.update(new)
        updates = {k: self._update_index(k, v) for k, v in current.items()}
        updated = self.indices.copy()
        updated.update(updates)
        return aliased.Mapping(updated)

    def _update_index(self, key: str, indices):
        """Update a single indexing object based on user input."""
        if not isinstance(indices, datatypes.Indices):
            axis = self.axes[key]
            indices = axis(*iterables.whole(indices))
        if isinstance(indices, datatypes.Coordinates):
            unit = MKS.get_unit(unit=indices.unit)
            return indices.with_unit(unit)
        return indices

    def update_assumptions(self, constraints: typing.Mapping):
        """Update the observing assumptions based on user constraints."""
        current = self.assumptions.copy()
        new = {k: v for k, v in constraints.items() if k not in self.axes}
        current.update(new)
        updates = {k: self._update_assumption(v) for k, v in current.items()}
        updated = self.assumptions.copy()
        updated.update(updates)
        return aliased.Mapping(updated)

    def _update_assumption(self, scalar):
        """Update a single assumption from user input."""
        if isinstance(scalar, quantities.Scalar):
            unit = MKS.get_unit(unit=scalar.unit())
            return scalar.unit(unit)
        measured = quantities.measure(scalar)
        assumption = [self._update_assumption(v) for v in measured[:]]
        return assumption[0] if len(assumption) == 1 else assumption

    def apply(self, constraints: typing.Mapping):
        """Construct the target variable within the given constraints."""
        indices = self.update_indices(constraints)
        assumptions = self.update_assumptions(constraints)
        application = Application(
            indices,
            assumptions,
            self.observables,
            self.reference,
        )
        self._result = application.evaluate(self.implementation)
        self._context = {
            'indices': aliased.Mapping(
                {
                    (k, *indices.alias(k, include=True)): v
                    for k, v in indices.items()
                    if k in self._result.axes
                }
            ),
            'assumptions': aliased.Mapping(
                {
                    (k, *assumptions.alias(k, include=True)): v
                    for k, v in assumptions.items()
                    if k in list(self.assumptions) + ['radius']
                }
            ),
        }

    @property
    def result(self):
        return self._result

    @property
    def context(self):
        return self._context

    def __str__(self) -> str:
        """A simplified representation of this object."""
        pairs = [
            f"implementation={self.implementation}",
            f"axes={self.axes}",
        ]
        return ', '.join(pairs)


class Observables(iterables.MappingBase):
    """An aliased mapping of observable quantities from an EPREM simulation."""

    def __init__(
        self,
        dataset: datasets.Dataset,
        arguments: parameters.Arguments,
    ) -> None:
        self.variables = dataset.variables
        self.functions = functions.Functions(dataset, arguments)
        vkeys = self.variables.keys
        fkeys = self.functions.keys
        self.primary = tuple(vkeys())
        self.derived = tuple(fkeys())
        self.names = self.primary + self.derived
        super().__init__(self.names)
        akeys = tuple(vkeys(aliased=True)) + tuple(fkeys(aliased=True))
        self.aliases = aliased.KeyMap(akeys)
        self.dataset = dataset
        self.arguments = arguments
        self._cache = {}

    def __getitem__(self, key: str):
        """Create the requested observable, if possible."""
        interface = self._implement(key)
        if interface is None:
            raise KeyError(f"No observable corresponding to {key!r}") from None
        return base.Observable(interface, self.aliases.get(key, key))

    def _implement(self, key: str):
        """Create an interface to this observable, if possible."""
        if key in self._cache:
            cached = self._cache[key]
            return Interface(
                cached['implementation'],
                self.dataset,
                dependencies=cached['dependencies'],
            )
        implementation = self.get_implementation(key)
        if implementation is not None:
            dependencies = self.get_dependencies(key)
            self._cache[key] = {
                'implementation': implementation,
                'dependencies': dependencies,
            }
            return Interface(
                implementation,
                self.dataset,
                dependencies=dependencies,
            )

    def get_implementation(self, key: str):
        """Get the implementation of the target observable."""
        if key in self.variables:
            return self.variables[key]
        if key in self.functions:
            return self.functions[key]
        if '/' in key or '*' in key:
            expression = algebra.Expression(key)
            return self._build_compound_observable(expression)

    def _build_compound_observable(self, expression: algebra.Expression):
        """Build a compound observable from an algebraic expression."""
        unique = list(
            {
                axis
                for term in expression
                for axis in self.get_implementation(term.base).axes
            }
        )
        canonical = ('time', 'shell', 'species', 'energy', 'mu')
        indices = (unique.index(axis) for axis in canonical if axis in unique)
        axes = tuple(unique[i] for i in indices)
        name = str(expression)
        return Compound(expression=expression, axes=axes, name=name)

    _RT = typing.TypeVar('_RT', bound=dict)
    _RT = typing.Dict[aliased.MappingKey, Dependency]

    def get_dependencies(self, key: str) -> _RT:
        """Get the dependencies for the given observable, if possible."""
        if key in self.variables:
            k = self.variables.alias(key, include=True)
            v = self.variables[key]
            return {k: v}
        if key in self.functions:
            function = self.functions[key]
            items = [
                self._get_dependency(name)
                for name in function.dependencies
            ]
            return {i[0]: i[1] for i in items}
        if '/' in key or '*' in key:
            terms = algebra.Expression(key).terms
            out = {}
            for term in terms:
                name = term.base
                if name in self.functions:
                    k = self.functions.alias(name, include=True)
                    out[k] = self.functions[name]
                out.update(self.get_dependencies(name))
            return out

    def _get_dependency(self, name: str):
        """Get an aliased dependency based on type."""
        if name in self.variables:
            k = self.variables.alias(name, include=True)
            v = self.variables[name]
            return k, v
        if name in self.functions:
            k = self.functions.alias(name, include=True)
            v = self.functions[name]
            return k, v
        if name in self.arguments:
            k = self.arguments.alias(name, include=True)
            v = self.arguments[name]
            return k, v




import typing

import numpy as np

from goats.core import aliased
from goats.core import datasets
from goats.core import datatypes
from goats.core import functions
from goats.core import iterables
from goats.core import observables
from goats.core import physical
from goats.core import quantities
from goats.eprem import parameters


class Function(iterables.ReprStrMixin):
    """A function from variables and scalars to a single variable."""

    def __init__(
        self,
        method: functions.Method,
        quantity: str,
        axes: typing.Tuple[str],
        dependencies: typing.Iterable[str]=None,
        name: str=None,
    ) -> None:
        self.method = method
        self.quantity = quantity
        self.axes = axes
        self.parameters = tuple(self.method.parameters)
        self.dependencies = tuple(dependencies or ())
        self.name = name or '<anonymous>'

    Argument = typing.TypeVar(
        'Argument',
        datatypes.Variable,
        quantities.Scalar,
        typing.Iterable[quantities.Scalar],
    )
    Argument = typing.Union[
        datatypes.Variable,
        quantities.Scalar,
        typing.Iterable[quantities.Scalar],
    ]

    def __call__(
        self,
        arguments: typing.Mapping[str, Argument],
        unit: typing.Union[str, quantities.Unit],
    ) -> datatypes.Variable:
        """Build a variable by calling the instance method."""
        arrays = []
        floats = []
        known = [
            argument for key, argument in arguments.items()
            if key in self.parameters
        ]
        for arg in known:
            if isinstance(arg, datatypes.Variable):
                arrays.append(np.array(arg))
            elif isinstance(arg, quantities.Scalar):
                floats.append(float(arg))
            elif (
                isinstance(arg, typing.Iterable)
                and all(isinstance(a, quantities.Scalar) for a in arg)
            ): floats.extend([float(a) for a in arg])
        data = self.method(*arrays, *floats)
        return datatypes.Variable(
            data,
            quantities.Unit(unit),
            self.axes,
            name=self.name,
        )

    def __str__(self) -> str:
        """A simplified representation of this object."""
        attrs = [
            f"method={self.method}",
            f"quantity='{self.quantity}'",
            f"axes={self.axes}",
        ]
        return ', '.join(attrs)


class Functions(aliased.Mapping):
    """Functions from variables and scalars to a single variable."""

    def __init__(
        self,
        dataset: datasets.Dataset,
        arguments: parameters.Arguments,
    ) -> None:
        super().__init__(mapping=functions.METHODS)
        self.dataset = dataset
        self._primary = (
            *tuple(self.dataset.variables.keys()),
            *tuple(arguments.keys()),
        )
        self._axes_cache = {}
        self._dependencies_cache = {}

    def __getitem__(self, key: str):
        """Construct the requested function object, if possible"""
        if method := self.get_method(key):
            axes = self.get_axes(key)
            quantity = observables.METADATA.get(key, {}).get('quantity', None)
            dependencies = self.get_dependencies(key)
            return Function(
                method,
                quantity,
                axes,
                dependencies=dependencies,
                name=key,
            )
        raise KeyError(f"No function corresponding to {key!r}")

    def get_method(self, key: str) -> functions.Method:
        """Attempt to retrieve a method by name based on `key`."""
        try:
            method = super().__getitem__(key)
        except KeyError:
            method = None
        return method

    def get_axes(self, key: str):
        """Retrieve or compute the axes corresponding to `key`."""
        if key in self._axes_cache:
            return self._axes_cache[key]
        method = self.get_method(key)
        self._removed = list(method.metadata.get('removed', []))
        self._restored = list(method.metadata.get('restored', []))
        self._accumulated = []
        axes = self._gather_axes(method)
        self._axes_cache[key] = axes
        return axes

    def _gather_axes(self, target: functions.Method):
        """Recursively gather appropriate axes."""
        for parameter in target.parameters:
            if parameter in self.dataset.variables:
                axes = self.dataset.iter_axes(parameter)
                self._accumulated.extend(axes)
            elif method := self.get_method(parameter):
                self._removed.extend(method.metadata.get('removed', []))
                self._restored.extend(method.metadata.get('restored', []))
                self._accumulated.extend(self._gather_axes(method))
        unique = set(self._accumulated) - set(self._removed)
        return self.dataset.resolve_axes(unique)

    def get_dependencies(self, key: str):
        """Compute the names of all dependencies of `key`."""
        if key in self._dependencies_cache:
            return self._dependencies_cache[key]
        try:
            target = self.get_method(key)
            p = self._gather_dependencies(target)
        except KeyError:
            return set()
        else:
            self._dependencies_cache[key] = p
            return p

    def _gather_dependencies(self, target: functions.Method):
        """Recursively gather the names of the target method's dependencies."""
        resolved = []
        for parameter in target.parameters:
            if parameter in self._primary:
                resolved.append(parameter)
            elif parameter in self:
                resolved.append(parameter)
                method = self.get_method(parameter)
                resolved.extend(self._gather_dependencies(method))
        return set(resolved)





import collections
import collections.abc
import typing

from goats.core import iterables


Aliases = typing.TypeVar('Aliases')
Aliases = typing.Union[str, typing.Iterable[str]]

class MappingKey(iterables.ReprStrMixin, collections.abc.Set):
    """A mapping key with associated aliases."""

    __slots__ = ('_aliases')

    _builtin = (tuple, list, set)

    @classmethod
    def supports(cls, key: str):
        """True if `key` can instantiate this class."""
        try:
            cls(key)
        except TypeError:
            return False
        return True

    def __init__(self, *a: Aliases) -> None:
        if not a:
            raise TypeError("At least one alias is required") from None
        self._aliases = self._from_iterable(a)

    @classmethod
    def _from_iterable(cls, it):
        try:
            length = len(it)
        except TypeError:
            length = None
        if length == 1 and not isinstance(it[0], str):
            return set(it[0])
        return set(it)

    def __iter__(self) -> typing.Iterator:
        return iter(self._aliases)

    def __len__(self) -> int:
        return len(self._aliases)

    def __contains__(self, key: str) -> bool:
        return key in self._aliases

    def __hash__(self) -> int:
        """Compute the hash of the underlying key set."""
        return hash(tuple(self._aliases))

    def _implement(operator):
        def method(self: 'MappingKey', other):
            return operator(self, MappingKey(other))
        def wrapper(self, other):
            result = method(self, other)
            if isinstance(result, typing.Iterable):
                return type(self)(result)
            return result
        return wrapper

    __eq__ = _implement(collections.abc.Set.__eq__)
    __and__ = _implement(collections.abc.Set.__and__)
    isdisjoint = _implement(collections.abc.Set.isdisjoint)
    __or__ = _implement(collections.abc.Set.__or__)
    __sub__ = _implement(collections.abc.Set.__sub__)
    __rsub__ = _implement(collections.abc.Set.__rsub__)
    __xor__ = _implement(collections.abc.Set.__xor__)

    def __str__(self) -> str:
        """A simplified representation of this instance."""
        return ' | '.join(str(alias) for alias in self._aliases)


_VT = typing.TypeVar('_VT')
class Mapping(collections.abc.Mapping):
    """A mapping class that supports aliased keys.
    
    Parameters
    ----------
    mapping : mapping
        An object that maps strings or iterables of strings to values of any
        type.

    Examples
    --------
    Create an instance from a standard `dict` with strings or tuples of strings
    as keys.

    >>> mapping = {'a': 1, ('b', 'B'): 2, ('c', 'C', 'c0'): -9}
    >>> amap = aliased.Mapping(mapping)
    >>> amap
    Mapping('a': 1, 'b | B': 2, 'c | C | c0': -9)

    The caller may access an individual item by any one (but only one) of its
    valid aliases.

    >>> amap['a']
    1
    >>> amap['b']
    2
    >>> amap['B']
    2
    >>> amap[('b', 'B')]
    ...
    KeyError: ('b', 'B')

    Aliased items are identical.

    >>> amap['b'] is amap['B']
    True

    The representation of keys, values, and items reflect the internal grouping
    by alias, but iterating over each produces de-aliased members. This behavior
    naturally supports loops and comprehensions, since access is only valid for
    individual items.

    >>> amap.keys()
    AliasedKeys(['a', 'b | B', 'c | C | c0'])
    >>> amap.values()
    AliasedValues([1, 2, -9])
    >>> amap.items()
    AliasedItems([('a', 1), ('b | B', 2), ('c | C | c0', -9)])
    >>> list(amap.keys())
    ['a', 'b', 'B', 'c', 'C', 'c0']
    >>> list(amap.values())
    [1, 2, 2, -9, -9, -9]
    >>> list(amap.items())
    [('a', 1), ('b', 2), ('B', 2), ('c', -9), ('C', -9), ('c0', -9)]
    >>> for k, v in amap.items():
    ...     print(k, amap[k], v)
    ... 
    a 1 1
    b 2 2
    B 2 2
    c -9 -9
    C -9 -9
    c0 -9 -9

    It is always possible to access the equivalent de-aliased `dict`.

    >>> amap.flat
    {'a': 1, 'b': 2, 'B': 2, 'c': -9, 'C': -9, 'c0': -9}

    Updates and deletions apply to all associated aliases.

    >>> amap['c'] = 5.6
    >>> list(amap.items())
    [('a', 1), ('b', 2), ('B', 2), ('c', 5.6), ('C', 5.6), ('c0', 5.6)]
    >>> del amap['c']
    >>> list(amap.items())
    [('a', 1), ('b', 2), ('B', 2)]

    Users may access all aliases for a given key via the
    `alias` method. Attempting to register an alias will raise a `KeyError` if
    it is already an alias for a different key.

    >>> amap.alias('b')
    MappingKey('B')
    >>> amap.alias('b', include=True)
    MappingKey('b | B')
    >>> amap.alias(b='my B')
    >>> amap.alias('b')
    MappingKey('B | my B')
    >>> amap
    Mapping('a': 1, 'b | my B | B': 2)
    >>> amap.alias(a='b')
    ...
    KeyError: "'b' is already an alias for '(B, my B)'"

    Notes
    -----
    The length of this object, as well as its keys, values, and items, is equal
    to the number of valid aliases it contains. This is consistent with the
    many-to-one nature of the mapping despite the fact that it internally stores
    aliases and values in a one-to-one mapping.

    """

    Aliasable = typing.TypeVar('Aliasable', bound=typing.Mapping)
    Aliasable = typing.Union[
        typing.Mapping[Aliases, _VT],
        typing.Mapping[str, typing.Mapping[str, typing.Any]],
    ]

    def __init__(
        self,
        mapping: typing.Union[Aliasable, 'Mapping']=None,
        aliases: str='aliases',
    ) -> None:
        """Initialize this instance."""
        self._aliased = self._build_aliased(mapping, aliases=aliases)
        self._keymap = {
            alias: key for key in self._aliased for alias in key
        }

    def _flat_keys(self) -> typing.KeysView[str]:
        """Define a flat list of all the keys in this mapping."""
        flattened = [key for keys in self._aliased.keys() for key in keys]
        return collections.abc.KeysView(flattened)

    def _build_aliased(
        self,
        mapping: typing.Mapping=None,
        aliases: str='aliases',
    ) -> dict:
        """Build the internal mapping of aliased items."""
        # Is it empty?
        if not mapping:
            return {}
        # Is it already an instance of this class?
        if isinstance(mapping, Mapping):
            return dict(mapping.items(aliased=True))
        # Does it have the form {<key>: {key: <aliases>, <k>: <value>, ...}}?
        string_keys = all(isinstance(key, str) for key in mapping)
        dict_values = all(isinstance(value, dict) for value in mapping.values())
        if string_keys and dict_values:
            return self._build_from_key(mapping, key=aliases)
        # Does it have the form {<aliased key>: <value>}?
        if all(MappingKey.supports(key) for key in mapping):
            return self._build_from_aliases(mapping)
        # Is it a built-in dictionary?
        if isinstance(mapping, dict):
            return mapping.copy()

    @property
    def flat(self) -> typing.Dict[str, _VT]:
        """Expand aliased items into a standard dictionary."""
        return {key: self[key] for key in self._flat_keys()}

    def __iter__(self) -> typing.Iterator:
        yield from self._flat_keys()

    def __len__(self) -> int:
        return len(self._flat_keys())

    def __getitem__(self, key: typing.Union[str, MappingKey]) -> _VT:
        """Look up a value by one of its keys."""
        if resolved := self._resolve(key):
            return self._aliased[resolved]
        raise KeyError(f"'{key!r}' is not a known name or alias.") from None

    def _resolve(self, key: typing.Union[MappingKey, typing.Any]):
        """Resolve `key` into an existing or new aliased key."""
        if isinstance(key, MappingKey):
            return self._look_up_key(key)
        return self._keymap.get(key)

    def _look_up_key(self, target: MappingKey):
        """Find the aliased key equivalent to `target`.
        
        Notes
        -----
        Checking ``key in self._aliased`` doesn't always work because `dict`
        look-up will first compare the key's hash value to those of existing
        keys before comparing the key itself. This may fail in the first stage
        due to the fact that `MappingKey.__hash__` uses `tuple.__hash__`, which
        depends on order, despite the fact that the second stage should pass
        because `MappingKey.__eq__` does not depend on order. See
        https://stackoverflow.com/q/327311/4739101.
        """
        try:
            found = next(
                key for key in self._aliased
                if key == target
            )
        except StopIteration:
            return
        else:
            return found

    def _build_from_aliases(
        self,
        mapping: typing.Union[Aliasable, 'Mapping'],
    ) -> typing.Dict[MappingKey, _VT]:
        """Build a `dict` that maps aliased keys to user values."""
        out = {}
        for key, value in mapping.items():
            try:
                aliased_key = next(k for k in out if key in k)
            except StopIteration:
                aliased_key = MappingKey(key)
            out[aliased_key] = value
        return out

    def _build_from_key(
        self,
        mapping: typing.Mapping[str, typing.Mapping[str, typing.Any]],
        key: str='aliases',
    ) -> typing.Dict[MappingKey, _VT]:
        """Build a `dict` with aliased keys taken from interior mappings.
        
        Parameters
        ----------
        mapping
            An object that maps string keys to interior mappings of strings to
            any type.

        key : str, default='aliases'
            The key in the interior mappings whose values to use as aliases for
            the corresponding keys in the user-provided mapping. Absence of this
            key from a given interior mapping simply means the corresponding key
            in the user-provided mapping will be the only alias for that entry.

        Examples
        --------
        Create aliased mappings from a user dictionary with the default alias
        key::

        >>> mapping = {
        ...     'a': {'aliases': ('A', 'a0'), 'n': 1, 'm': 'foo'},
        ...     'b': {'aliases': 'B', 'n': -4},
        ... }
        >>> amap = aliased.Mapping(mapping)
        >>> amap
        aliased.Mapping('a0 | A | a': {'n': 1, 'm': 'foo'}, 'b | B': {'n': -4})
        >>> amap['a']
        {'n': 1, 'm': 'foo'}

        Create aliased mappings from a user dictionary, swapping the alias key::

        >>> mapping = {
        ...     'a': {'foo': 'A', 'bar': 'a0'},
        ...     'b': {'foo': 'B', 'bar': 'b0'},
        ... }
        >>> amap = aliased.Mapping(mapping, aliases='foo')
        >>> amap
        aliased.Mapping('A | a': a0, 'b | B': b0)
        >>> amap['a']
        'a0'
        >>> amap = aliased.Mapping(mapping, aliases='bar')
        >>> amap
        aliased.Mapping('a0 | a': A, 'b | b0': B)
        >>> amap['a']
        'A'
        """
        keys = self.extract_keys(mapping, aliases=key)
        values = [
            {k: v for k, v in group.items() if k != key}
            for group in mapping.values()
        ]
        return dict(zip(keys, values))

    @classmethod
    def fromkeys(
        cls,
        mapping: typing.Mapping[str, typing.Mapping[str, typing.Any]],
        aliases: str='aliases',
        value: typing.Any=None,
    ): # How do I annotate this so it's correct for subclasses?
        """Create an aliased mapping based on another mapping's keys.

        Parameters
        ----------
        user : mapping
            Same as for `~aliases.Mapping.__init__`.

        aliases : string
            Same as for `~aliases.Mapping.__init__`.

        value : any
            The fill value to use for all items.

        Returns
        -------
        aliased mapping
            A new instance of this class, with aliased keys taken from the
            user-provided mapping and each value set to the given value.

        Examples
        --------
        Create an aliased mapping with keys taken from a pre-defined dictionary
        and all values set to -1.0

        >>> mapping = {
        ...     'a': {'aliases': ('A', 'a0'), 'n': 1, 'm': 'foo'},
        ...     'b': {'aliases': 'B', 'n': -4},
        ... }
        >>> amap = aliased.Mapping.fromkeys(mapping, value=-1.0)
        >>> amap
        aliased.Mapping('a0 | a | A': -1.0, 'B | b': -1.0)
        """
        keys = cls.extract_keys(mapping, aliases=aliases)
        d = {k: value for k in keys}
        return cls(d)

    @classmethod
    def extract_keys(
        cls,
        mapping: typing.Mapping[str, typing.Mapping[str, typing.Any]],
        aliases: str='aliases',
    ) -> typing.List[MappingKey]:
        """Extract keys for use in an aliased mapping.
        
        Parameters
        ----------
        mapping
            Same as for `~aliased.Mapping.__init__`.

        aliases : string
            Same as for `~aliased.Mapping.__init__`.

        Examples
        --------
        >>> mapping = {
        ...     'a': {'aliases': ('A', 'a0'), 'n': 1, 'm': 'foo'},
        ...     'b': {'aliases': 'B', 'n': -4},
        ... }
        >>> keys = Mapping.extract_keys(mapping)
        >>> keys
        [MappingKey('a | A | a0'), MappingKey('B | b')]
        """
        if isinstance(mapping, Mapping):
            return mapping.keys(aliased=True)
        return [
            MappingKey(k) | MappingKey(v.get(aliases, ()))
            for k, v in mapping.items()
        ]

    def alias(self, *current, include=False):
        """Get the alias for an existing key."""
        if current:
            if len(current) > 1:
                raise ValueError(
                    "Can only get one aliased key at a time"
                ) from None
            key = current[0]
            if include:
                return self._resolve(key)
            return self._resolve(key) - key

    def __eq__(self, other: typing.Mapping) -> bool:
        """Define equality between this and another object."""
        if not isinstance(other, typing.Mapping):
            return NotImplemented
        if isinstance(other, Mapping):
            return self.items() == other.items()
        return dict(self.items()) == dict(other.items())

    def __or__(self, other: 'Mapping'):
        """Merge this aliased mapping with another."""
        items = dict((*self.items(aliased=True), *other.items(aliased=True)))
        return type(self)(items)

    def __str__(self) -> str:
        """A simplified representation of this instance."""
        return ', '.join(
            f"'{g}': {v!r}"
            for g, v in zip(self._aliased.keys(), self._aliased.values())
        )

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        module = f"{self.__module__.replace('goats.', '')}."
        name = self.__class__.__qualname__
        return f"{module}{name}({self})"

    def keys(self, aliased: bool=False):
        """A view on this instance's keys."""
        return KeysView(self, aliased=aliased)

    def values(self, aliased: bool=False):
        """A view on this instance's values."""
        return ValuesView(self, aliased=aliased)

    def items(self, aliased: bool=False):
        """A view on this instance's key-value pairs."""
        return ItemsView(self, aliased=aliased)

    def copy(self):
        """Create a shallow copy of this instance."""
        return type(self)(self._aliased)


class MappingView(iterables.ReprStrMixin, collections.abc.MappingView):
    """Base class for views of aliased mappings."""

    __slots__ = ('_mapping', '_keys')

    def __init__(self, mapping: Mapping, aliased: bool=False) -> None:
        super().__init__(mapping)
        aliases = mapping._aliased.keys()
        self._keys = aliases if aliased else mapping._flat_keys()
        self._mapping = mapping

    def __len__(self):
        """Called for len(self)."""
        return len(self._keys)

    def __str__(self):
        """A simplified representation of this object."""
        return str(list(self))


class KeysView(MappingView, collections.abc.KeysView):
    """A view on the keys of an aliased mapping."""

    def __iter__(self):
        """Iterate over aliased mapping keys."""
        yield from self._keys


class ValuesView(MappingView, collections.abc.ValuesView):
    """A view on the values of an aliased mapping."""

    def __iter__(self):
        """Iterate over aliased mapping values."""
        for key in self._keys:
            yield self._mapping[key]


class ItemsView(MappingView, collections.abc.ItemsView):
    """A view on the key-value pairs of an aliased mapping."""

    def __iter__(self):
        """Iterate over aliased mapping items."""
        for key in self._keys:
            yield (key, self._mapping[key])


class MutableMapping(Mapping, collections.abc.MutableMapping):
    """A mutable version of `Mapping`.
    
    Parameters
    ----------
    mapping : mapping
        An object that maps strings or iterables of strings to values of any
        type.

    Examples
    --------
    These examples build on the examples shown in `Mapping`.

    >>> mutable = iterables.MutableMapping(aliased)

    Updates and deletions apply to all associated aliases.

    >>> mutable['c'] = 5.6
    >>> list(mutable.items())
    [('a', 1), ('b', 2), ('B', 2), ('c', 5.6), ('C', 5.6), ('c0', 5.6)]
    >>> del mutable['c']
    >>> list(mutable.items())
    [('a', 1), ('b', 2), ('B', 2)]

    Users may access all aliases for a given key, or register new ones, via the
    `alias` method. Attempting to register an alias will raise a `KeyError` if
    it is already an alias for a different key.

    >>> mutable.alias('b')
    MappingKey('B')
    >>> mutable.alias('b', include=True)
    MappingKey('b | B')
    >>> mutable.alias(b='my B')
    >>> mutable.alias('b')
    MappingKey('B | my B')
    >>> mutable
    Mapping('a': 1, 'b | my B | B': 2)
    >>> mutable.alias(a='b')
    ...
    KeyError: "'b' is already an alias for '(B, my B)'"

    Notes
    -----
    See note on lengths at `~Mapping`.
    """

    def __setitem__(self, key: str, value: _VT):
        """Assign a value to `key` and its aliases."""
        resolved = self._resolve(key) or MappingKey(key)
        self._aliased[resolved] = value
        self._refresh()

    def __delitem__(self, key: str):
        """Remove the item corresponding to `key`."""
        resolved = self._resolve(key)
        if not resolved:
            raise KeyError(f"'{key!r}' is not a known name or alias.") from None
        del self._aliased[resolved]
        self._refresh()

    def _refresh(self):
        """Perform common tasks after setting or deleting an item."""
        self._keymap = {
            alias: key for key in self._aliased for alias in key
        }

    def alias(self, *current: str, include=False, **new: str):
        """Get the alias for an existing key or register new ones."""
        if current and new:
            raise TypeError(
                "Can't get and set aliases at the same time"
            ) from None
        if current:
            if len(current) > 1:
                raise ValueError(
                    "Can only get one aliased key at a time"
                ) from None
            key = current[0]
            if include:
                return self._resolve(key)
            return self._resolve(key) - key
        for key, alias in new.items():
            if alias:
                if alias in self._flat_keys():
                    raise self._not_available(alias) from None
                updated = self._resolve(key) | alias
                self._aliased[updated] = self[key]
                del self[key]

    def _not_available(self, key: str) -> typing.NoReturn:
        """True if this key is not currently in use."""
        aliases = self.alias(key)
        this = ", ".join(f'{a}' for a in aliases)
        if len(aliases) > 1:
            this = f"({this})"
        return KeyError(f"'{key}' is already an alias for {this!r}")


class NameMap(iterables.MappingBase):
    """A mapping from aliases to canonical names.
    
    Notes
    -----
    See note on lengths at `~Mapping`.
    """

    AliasDefinitions = typing.TypeVar('AliasDefinitions')
    AliasDefinitions = typing.Union[
        typing.Iterable[typing.Iterable[str]],
        typing.Mapping[str, typing.Iterable[str]],
        typing.Mapping[str, typing.Mapping[str, typing.Iterable[str]]],
    ]

    AliasReferences = typing.TypeVar('AliasReferences')
    AliasReferences = typing.Union[
        typing.Iterable[str],
        typing.Mapping[str, typing.Any],
    ]

    def __init__(
        self,
        defs: AliasDefinitions,
        refs: AliasReferences=None,
        key: str='aliases',
    ) -> None:
        names = self._get_names(defs, refs)
        self._mapping = self._build_mapping(names, defs, key)
        super().__init__(self._mapping.keys())
        self._init = {'refs': refs, 'defs': defs, 'key': key}

    RT = typing.TypeVar('RT')
    RT = typing.Union[
        typing.Iterable[str],
        typing.Iterable[typing.Iterable[str]],
    ]
    def _get_names(
        self,
        defs: AliasDefinitions,
        refs: AliasReferences=None,
    ) -> RT:
        """Create an iterable of canonical names, if possible."""
        if isinstance(refs, typing.Mapping):
            return refs.keys()
        if isinstance(refs, typing.Iterable):
            return refs
        if isinstance(defs, typing.Mapping):
            return defs.keys()
        raise TypeError(
            f"Can't create name map from {defs!r} and {refs!r}"
        ) from None

    def _build_mapping(self, names, aliases, key):
        """Build the internal mapping from aliases to canonical names.

        This method first creates an identity map of canonical
        names (i.e., `name` -> `name`), with which it spawns a trivial aliased
        mapping. It then determines the appropriate aliases for each canonical
        name and updates the aliased-mapping keys. The result is a mapping from
        one or more aliases to a canonical name. In case there are no aliases
        associated with a canonical name, its aliased key will simply contain
        itself.
        """
        identity = {name: name for name in names}
        namemap = MutableMapping(identity)
        updates = self._get_aliases(names, aliases, key)
        namemap.alias(**updates)
        return Mapping(namemap)

    def _get_aliases(self, names, defs: AliasDefinitions, key):
        """Determine the appropriate aliases for each canonical name."""
        # Mapping <: Iterable, so we need to check Mapping first.
        if isinstance(defs, typing.Mapping):
            # There are two allowed types of Mapping:
            # 1) Mapping[str, Mapping[str, Iterable[str]]]
            # 2) Mapping[str, Iterable[str]]
            
            # Make sure the keys are all strings.
            if any(k for k in defs if not isinstance(k, str)):
                raise TypeError("All aliases must be strings") from None
            # Again, we need to check Mapping values before Iterable values.
            if all(isinstance(v, typing.Mapping) for v in defs.values()):
                return {
                    name: tuple(v.get(key, ()))
                    for name, v in defs.items()
                }
            if all(isinstance(v, typing.Iterable) for v in defs.values()):
                return {name: tuple(aliases) for name, aliases in defs.items()}
        # Alias definitions are in a non-mapping iterable. We may want to
        # further check that each member of `defs` is itself an iterable of
        # strings.
        only_iterables = all(isinstance(d, typing.Iterable) for d in defs)
        if isinstance(defs, typing.Iterable) and only_iterables:
            return {
                name: tuple(aliases)
                for aliases in defs
                for name in names
                if name in aliases
            }
        return {}

    def invert(self):
        """Invert the mapping from {aliases -> name} to {name -> aliases}."""
        inverted = {
            name: list(aliases)
            for aliases, name in self._mapping.items(aliased=True)
        }
        self._mapping = Mapping(inverted)
        return self

    def __getitem__(self, key: str):
        """Get the canonical name for `key`."""
        if key in self._mapping:
            return self._mapping[key]
        raise KeyError(key)

    def __str__(self) -> str:
        """A simplified representation of this object."""
        items = {f"{k}: '{v}'" for k, v in self._mapping.items(aliased=True)}
        return ', '.join(items)

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return f"{self.__class__.__qualname__}({self})"

    def keys(self, aliased: bool=False):
        """A view on this object's aliased keys."""
        return Mapping.keys(self._mapping, aliased=aliased)

    def values(self, aliased: bool=False):
        """A view on this object's aliased values."""
        return Mapping.values(self._mapping, aliased=aliased)

    def items(self, aliased: bool=False):
        """A view on this object's aliased key-value pairs."""
        return Mapping.items(self._mapping, aliased=aliased)

    def copy(self):
        """Make a shallow copy of this object."""
        return type(self)(**self._init)


class KeyMap(iterables.MappingBase):
    """A collection that associates common aliases."""

    def __init__(self, __keys: typing.Iterable[Aliases]) -> None:
        """
        Parameters
        ----------
        __keys : iterable
            An iterable collection of associated keys. Each key may be a string
            or an iterable of strings (including instances of `~MappingKey`).
        """
        self._aliased = [MappingKey(key) for key in __keys]
        self._flat = [key for alias in self._aliased for key in alias]
        super().__init__(self._flat)

    def __getitem__(self, key: str) -> MappingKey:
        """Look up aliases for key."""
        try:
            found = next(entry for entry in self._aliased if key in entry)
        except StopIteration as err:
            raise KeyError(f"{key!r} not found") from err
        else:
            return found





"""Tools for managing datasets."""

import abc
import collections.abc
import numbers
import typing

import netCDF4
import numpy
import numpy.typing

from goats.core import datatypes
from goats.core import aliased
from goats.core import iotools
from goats.core import iterables
from goats.core import observables
from goats.core import quantities


class DataViewer(collections.abc.Mapping):
    """An abstract base class for data-viewing objects."""

    def __init__(self, path: iotools.ReadOnlyPath) -> None:
        self.members = self.get_members(path)

    def __iter__(self) -> typing.Iterator:
        return iter(self.members)

    def __len__(self) -> int:
        return len(self.members)

    @abc.abstractmethod
    def get_members(self, path: iotools.ReadOnlyPath) -> typing.Mapping:
        """Get the appropriate members for this viewer."""
        pass

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return '\n\n'.join(f"{k}:\n{v!r}" for k, v in self.items())

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return f"\n::{self.__class__.__qualname__}::\n\n{self}"


class DatasetVariable(typing.NamedTuple):
    """A dataset variable."""

    data: numpy.typing.ArrayLike
    unit: str
    axes: typing.Tuple[str]
    name: str

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self._display(sep='\n')

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        display = self._display(sep='\n', tab=4)
        return f"{self.__class__.__qualname__}(\n{display}\n)"

    def _display(
        self,
        sep: str=', ',
        tab: int=0,
    ) -> str:
        """Helper for `__str__` and `__repr__`."""
        attrs = [
            f"data={type(self.data)}",
            f"unit={self.unit!r}",
            f"axes={self.axes}",
            f"name={self.name!r}",
        ]
        indent = ' ' * tab
        return sep.join(f"{indent}{attr}" for attr in attrs)


class NetCDFVariables(DataViewer):
    """An object for viewing variables in a NetCDF dataset."""

    def get_members(self, path: iotools.ReadOnlyPath) -> typing.Mapping:
        dataset = netCDF4.Dataset(path, 'r')
        return dataset.variables

    def __getitem__(self, name: str) -> DatasetVariable:
        if name in self.members:
            data = self.members[name]
            unit = self._get_unit_from_data(data)
            axes = self._get_axes_from_data(data)
            return DatasetVariable(data, unit, axes, name)
        raise KeyError(f"No variable called '{name}'")

    def _get_axes_from_data(self, data):
        """Compute appropriate variable axes from a dataset object."""
        return tuple(getattr(data, 'dimensions', ()))

    def _get_unit_from_data(self, data):
        """Compute appropriate variable units from a dataset object."""
        available = (
            getattr(data, attr) for attr in ('unit', 'units')
            if hasattr(data, attr)
        )
        return next(available, None)


class DatasetAxis(typing.NamedTuple):
    """A dataset axis."""

    size: int
    name: str

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return f"{self.name!r}, size={self.size}"

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        module = f"{self.__module__.replace('goats.', '')}."
        name = self.__class__.__qualname__
        return f"{module}{name}({self})"


class NetCDFAxes(DataViewer):
    """An object for viewing axes in a NetCDF dataset."""

    def get_members(self, path: iotools.ReadOnlyPath) -> typing.Mapping:
        dataset = netCDF4.Dataset(path, 'r')
        return dataset.dimensions

    def __getitem__(self, name: str) -> DatasetAxis:
        if name in self.members:
            data = self.members[name]
            size = getattr(data, 'size', None)
            return DatasetAxis(size, name)
        raise KeyError(f"No axis corresponding to {name!r}")


class ViewerFactory(collections.abc.MutableMapping):
    """A class that creates appropriate viewers for a dataset."""

    _viewer_map = {
        '.nc': {
            'variables': NetCDFVariables,
            'axes': NetCDFAxes,
        }
    }

    def __init__(self, path: iotools.ReadOnlyPath) -> None:
        self._viewers = self._get_viewers(path)
        self.path = path

    def __iter__(self) -> typing.Iterator:
        return iter(self._viewers)

    def __len__(self) -> int:
        return len(self._viewers)

    def _get_viewers(
        self,
        path: iotools.ReadOnlyPath,
    ) -> typing.Dict[str, typing.Type[DataViewer]]:
        """Get the viewers for this file format.

        This may expand to accommodate additional file formats or alternate
        methods.
        """
        try:
            viewers = self._viewer_map[path.suffix]
        except KeyError:
            TypeError(f"Unrecognized file type for {path}")
        else:
            return viewers

    def __getitem__(self, group: str):
        """Get the appropriate viewer for this dataset group."""
        if group in self._viewers:
            viewer = self._viewers[group]
            return viewer(self.path)
        raise KeyError(f"No viewer for group '{group}' in {self.path}")

    def __setitem__(self, group: str, viewer: typing.Type[DataViewer]):
        """Associate a new viewer with this dataset group."""
        self._viewers[group] = viewer

    def __delitem__(self, group: str):
        """Delete the viewer associated with this dataset group."""
        del self._viewers[group]


class SubsetKeys(typing.NamedTuple):
    """A subset of names of attributes."""

    full: typing.Tuple[str]
    aliased: typing.Tuple[aliased.MappingKey]
    canonical: typing.Tuple[str]


class DatasetView(iterables.ReprStrMixin, metaclass=iotools.PathSet):
    """A format-agnostic view of a dataset.
    
    An instance of this class provides aliased access to variables and axes
    defined in a specific dataset, given a path to that dataset. It is designed
    to provide a single interface, regardless of file type, with as little
    overhead as possible. Therefore, it does not attempt to modify attributes
    (e.g., converting variable units), since doing so could result in reading a
    potentially large array from disk.
    """

    def __init__(self, path: iotools.PathLike) -> None:
        self.path = iotools.ReadOnlyPath(path)
        self.viewers = ViewerFactory(self.path)
        self._variables = None
        self._axes = None
        self._units = None

    @property
    def variables(self):
        """The variables in this dataset."""
        if self._variables is None:
            variables = {
                observables.ALIASES.get(name, name): variable
                for name, variable in self.viewers['variables'].items()
            }
            self._variables = aliased.Mapping(variables)
        return self._variables

    @property
    def units(self):
        """The unit of each variable, if available."""
        if self._units is None:
            units = {
                name: variable.unit
                for name, variable in self.variables.items(aliased=True)
            }
            self._units = aliased.Mapping(units)
        return self._units

    @property
    def axes(self):
        """The axes in this dataset."""
        if self._axes is None:
            axes = {
                observables.ALIASES.get(name, name): axis
                for name, axis in self.viewers['axes'].items()
            }
            self._axes = aliased.Mapping(axes)
        return self._axes

    def available(self, key: str):
        """Provide the names of available attributes."""
        if key in {'variable', 'variables'}:
            return SubsetKeys(
                full=tuple(self.variables),
                aliased=tuple(self.variables.keys(aliased=True)),
                canonical=tuple(self.viewers['variables'].keys()),
            )
        if key in {'axis', 'axes'}:
            return SubsetKeys(
                full=tuple(self.axes),
                aliased=tuple(self.axes.keys(aliased=True)),
                canonical=tuple(self.viewers['axes'].keys()),
            )

    def use(self, **viewers) -> 'DatasetView':
        """Update the viewers for this instance."""
        self.viewers.update(viewers)
        return self

    def __str__(self) -> str:
        return str(self.path)


S = typing.TypeVar('S', bound='Variables')


class Variables(aliased.Mapping):
    """An interface to dataset variables.
    
    This class provides aliased key-based access to all variables in a dataset.
    It converts each requested dataset variable into a `~quantities.Variable`
    instance with the appropriate MKS unit.
    """

    def __init__(self, dataset: DatasetView) -> None:
        known = {
            k: v for k, v in dataset.variables.items(aliased=True)
            if k in observables.METADATA
        }
        super().__init__(known)
        self._system = quantities.MetricSystem('mks')
        self._units = None
        self._cache = {}

    @property
    def units(self):
        """The MKS unit of each variable."""
        if self._units is None:
            units = {
                name: self._get_unit(name)
                for name in self.keys(aliased=True)
            }
            self._units = aliased.Mapping(units)
        return self._units

    def _get_unit(self, name: str):
        """Get a standard unit for the named variable."""
        metric = self._system[observables.METADATA[name]['quantity']]
        return quantities.Unit(metric.unit)

    def __getitem__(self, key: str):
        """Create the named variable, if possible."""
        if key in self._cache:
            return self._cache[key]
        datavar = super().__getitem__(key)
        variable = datatypes.Variable(
            datavar.data,
            standardize(datavar.unit),
            datavar.axes,
            name=observables.ALIASES[key],
        )
        result = variable.convert_to(self.units[key])
        self._cache[key] = result
        return result


class Indexers(aliased.Mapping):
    """The default collection of axis indexers."""

    def __init__(self, dataset: DatasetView) -> None:
        self.dataset = dataset
        self.references = {}
        super().__init__(dataset.axes)

    def __getitem__(self, key: str) -> datatypes.Indexer:
        """Get the default indexer for `key`."""
        axis = super().__getitem__(key)
        reference = range(axis.size)
        method = lambda _: datatypes.Indices(reference)
        return datatypes.Indexer(method, reference)


class Axes(aliased.Mapping):
    """An interface to dataset axes."""

    def __init__(
        self,
        dataset: DatasetView,
        factory: typing.Type[Indexers],
    ) -> None:
        indexers = factory(dataset)
        super().__init__(indexers)
        self.dataset = dataset

    def __getitem__(self, key: str) -> datatypes.Axis:
        indexer = super().__getitem__(key)
        size = self.dataset.axes[key].size
        name = f"'{observables.ALIASES.get(key, key)}'"
        return datatypes.Axis(size, indexer, name=name)


class Dataset:
    """The user interface to a dataset."""

    def __init__(
        self,
        path: iotools.PathLike,
        indexers: typing.Type[Indexers]=None,
    ) -> None:
        self.path = path
        self.view = DatasetView(path)
        self.indexers = indexers or Indexers
        self._variables = None
        self._axes = None

    @property
    def variables(self):
        """Objects representing the variables in this dataset."""
        if self._variables is None:
            self._variables = Variables(self.view)
        return self._variables

    @property
    def axes(self):
        """Objects representing the axes in this dataset."""
        if self._axes is None:
            self._axes = Axes(self.view, self.indexers)
        return self._axes

    Default = typing.TypeVar('Default')

    def iter_axes(self, name: str, default: Default=None):
        """Iterate over the axes for the named variable."""
        if name in self.variables:
            return iter(self.variables[name].axes)
        if default is not None:
            return default
        raise ValueError(
            f"Can't iterate over axes of missing variable {name!r}"
        ) from None

    def resolve_axes(self, names: typing.Iterable[str]):
        """Compute and order the available axes in `names`."""
        axes = self.view.available('axes').canonical
        return tuple(name for name in axes if name in names)


substitutions = {
    'julian date': 'day',
    'shell': '1',
    'cos(mu)': '1',
    'e-': 'e',
    '# / cm^2 s sr MeV': '# / cm^2 s sr MeV/nuc',
}
"""Conversions from non-standard units."""

T = typing.TypeVar('T')

def standardize(unit: T):
    """Replace this unit string with a standard unit string, if possible.

    This function looks for `unit` in the known conversions and returns the
    standard unit string if it exists. If this doesn't find a standard unit
    string, it just returns the input.
    """
    unit = substitutions.get(str(unit), unit)
    if '/' in unit:
        num, den = str(unit).split('/', maxsplit=1)
        unit = ' / '.join((num.strip(), f"({den.strip()})"))
    return unit





import collections.abc
import numbers
import typing

import numpy as np


class NumericalSequence(collections.abc.Sequence):
    """Abstract base class representing a sequence of numerical values.

    This ABC is a partial implementation of `collections.abc.Sequence`, indended
    to serve as a basis for creating array-like objects that interoperate with
    `numpy` arrays. It defines the required `__len__` method but does not define
    the required `__getitem__` method.
    """
    def __init__(self, values: typing.Iterable[numbers.Number]) -> None:
        self.values = values

    def __len__(self) -> int:
        return len(self.values)

    def __array__(self, *args, **kwargs) -> np.ndarray:
        """Support casting to `numpy` array types."""
        return np.asarray(self.values, *args, **kwargs)


class Array(NumericalSequence, np.lib.mixins.NDArrayOperatorsMixin):
    """An indexable numerical sequence."""

    def __init__(
        self,
        values: typing.Iterable[numbers.Number],
        axes: typing.Iterable[str],
    ) -> None:
        super().__init__(values)
        self._axes = axes
        self._data = None

    @property
    def axes(self):
        """The names of this array's indexable axes."""
        return tuple(self._axes)

    @property
    def naxes(self):
        """The number of indexable axes in this array."""
        return len(self.axes)

    @property
    def data(self) -> np.ndarray:
        """The `numpy` representation of this array's values."""
        if self._data is None:
            # FIXME: This is not a rigorous solution to the problem of certain
            # types of `self.values` not producing an array.
            try:
                self._data = np.asarray(self.values)
            except ValueError:
                self._data = np.asarray(self.values[:])
        return self._data

    def __getitem__(self, indices):
        """Access array values via index or slice notation."""
        idx = np.index_exp[indices]
        return self.data[idx]

    _HANDLED_TYPES = (np.ndarray, numbers.Number, list)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Provide support for `numpy` universal functions.

        See https://numpy.org/doc/stable/reference/arrays.classes.html for more
        information on use of this special method.

        See
        https://numpy.org/doc/stable/reference/generated/numpy.lib.mixins.NDArrayOperatorsMixin.html#numpy.lib.mixins.NDArrayOperatorsMixin
        for the specific implementation example that this class follows.

        This method first ensures that the input types (as well as the type of
        `out`, if given) are supported types.
        """
        out = kwargs.get('out', ())
        for x in inputs + out:
            if not isinstance(x, self._HANDLED_TYPES + (Array,)):
                return NotImplemented
        inputs = tuple(x.values if isinstance(x, Array) else x for x in inputs)
        if out:
            kwargs['out'] = tuple(
                x.values if isinstance(x, Array) else x for x in out
            )
        result = getattr(ufunc, method)(*inputs, **kwargs)
        if type(result) is tuple:
            return tuple(self._new_from_func(x) for x in result)
        if method == 'at':
            return None
        return self._new_from_func(result)

    _HANDLED_FUNCTIONS = {}

    def __array_function__(self, func, types, args, kwargs):
        """Provide support for functions in the `numpy` public API.

        See https://numpy.org/doc/stable/reference/arrays.classes.html for more
        information of use of this special method. The implementation shown here
        is a combination of the example on that page and code from the
        definition of `EncapsulateNDArray.__array_function__` in
        https://github.com/dask/dask/blob/main/dask/array/tests/test_dispatch.py

        The initial `issubclass` check allows subclasses that don't override
        `__array_function__` to handle `Array` objects.
        """
        accepted = (type(self), np.ndarray) + np.ScalarType
        if not all(issubclass(ti, accepted) for ti in types):
            return NotImplemented
        if func in self._HANDLED_FUNCTIONS:
            arr = self._HANDLED_FUNCTIONS[func](*args, **kwargs)
            return self._new_from_func(arr)
        args = tuple(
            arg
            if not isinstance(arg, type(self)) else arg.values
            for arg in args
        )
        types = tuple(ti for ti in types if not issubclass(ti, type(self)))
        values = self.__array__()
        arr = values.__array_function__(func, types, args, kwargs)
        return self._new_from_func(arr)

    def _new_from_func(self, result):
        """Create a new instance from the result of a `numpy` function."""
        if not isinstance(result, np.ndarray):
            return result
        return Array(result, self.axes)

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return np.array2string(self.data, separator=', ')

    def __repr__(self) -> str:
        """An unambiguous representation of this object.

        See
        https://numpy.org/doc/stable/reference/generated/numpy.array2string.html#numpy.array2string
        for `array2string` options.
        """
        name = self.__class__.__qualname__
        prefix = f"{name}("; suffix = ")"
        string = np.array2string(
            self.data, prefix=prefix, suffix=suffix, separator=', ',
        )
        indent = ' ' * len(prefix)
        return f"{name}{self.axes}(\n{indent}{string}\n)"

    @classmethod
    def implements(cls, numpy_function):
        """Register an `__array_function__` implementation for Array.

        See https://numpy.org/doc/stable/reference/arrays.classes.html for the
        suggestion on which this method is based.

        EXAMPLE
        -------
        Override `numpy.mean` with a version that accepts no keyword arguments::

            @Array.implements(np.mean)
            def mean(a: Array, **kwargs) -> Array:
                if kwargs:
                    msg = "Cannot pass keywords to numpy.mean with Array"
                    raise TypeError(msg)
                return np.sum(a) / len(a)

        This will compute the mean of the underlying values when called with no
        arguments, but will raise an exception when called with arguments:

            >>> v = Array([[1, 2], [3, 4]])
            >>> np.mean(v)
            5.0
            >>> np.mean(v, axis=0)
            ...
            TypeError: Cannot pass keywords to numpy.mean with Array

        """
        def decorator(func):
            cls._HANDLED_FUNCTIONS[numpy_function] = func
            return func
        return decorator


@Array.implements(np.mean)
def _array_mean(a: Array, **kwargs):
    """Compute the mean and update array dimensions, if necessary."""
    values = a.data.mean(**kwargs)
    if (axis := kwargs.get('axis')) is not None:
        a._axes = tuple(
            d for d in a.axes
            if a.axes.index(d) != axis
        )
    return values


